import numpy as np

import openmdao.api as om


class WaterEngine(om.Group):
    """
    Computes thrust and water flow for a water.

    Simplifications:
     - the pressure due to the water column in the non inertial frame (i.e.
       under a+g acceleration) is insignificant compared to the air pressure
     - the water does not have appreciable speed inside the bottle
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='water_exhaust_speed',
                           subsys=_WaterExhaustSpeed(num_nodes=nn),
                           promotes=['p', 'p_a', 'rho_w'])

        self.add_subsystem(name='water_flow_rate',
                           subsys=_WaterFlowRate(num_nodes=nn),
                           promotes=['A_out', 'Vdot'])

        self.add_subsystem(name='pressure_rate',
                           subsys=_PressureRate(num_nodes=nn),
                           promotes=['p', 'k', 'V_b', 'V_w', 'Vdot', 'pdot'])

        self.add_subsystem(name='water_thrust',
                           subsys=_WaterThrust(num_nodes=nn),
                           promotes=['rho_w', 'A_out', 'F'])

        self.set_input_defaults('A_out', val=np.pi*13e-3**2/4*np.ones(nn))
        self.set_input_defaults('p', val=6.5e5*np.ones(nn))

        self.connect('water_exhaust_speed.v_out', 'water_flow_rate.v_out')
        self.connect('water_exhaust_speed.v_out', 'water_thrust.v_out')


class _WaterExhaustSpeed(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(name='rho_w', val=1e3*np.ones(nn), desc='water density', units='kg/m**3')
        self.add_input(name='p', val=6.5e5*np.ones(nn), desc='air pressure', units='N/m**2')  # 5.5bar = 80 psi
        self.add_input(name='p_a', val=1.01e5*np.ones(nn), desc='air pressure', units='N/m**2')

        self.add_output(name='v_out', shape=(nn,), desc='water exhaust speed', units='m/s')

        ar = np.arange(nn)

        self.declare_partials(of='*', wrt='*', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        p = inputs['p']
        p_a = inputs['p_a']
        rho_w = inputs['rho_w']

        outputs['v_out'] = np.sqrt(2*(p-p_a)/rho_w)

    def compute_partials(self, inputs, partials):
        p = inputs['p']
        p_a = inputs['p_a']
        rho_w = inputs['rho_w']

        v_out = np.sqrt(2*(p-p_a)/rho_w)

        partials['v_out', 'p'] = 1/v_out/rho_w
        partials['v_out', 'p_a'] = -1/v_out/rho_w
        partials['v_out', 'rho_w'] = 1/v_out*(-(p-p_a)/rho_w**2)


class _WaterFlowRate(om.ExplicitComponent):
    """ Computer water flow rate"""
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(name='A_out', val=np.ones(nn), desc='nozzle outlet area', units='m**2')
        self.add_input(name='v_out', val=np.zeros(nn), desc='water exhaust speed', units='m/s')

        self.add_output(name='Vdot', shape=(nn,), desc='water flow', units='m**3/s')

        ar = np.arange(nn)

        self.declare_partials(of='*', wrt='*', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        A_out = inputs['A_out']
        v_out = inputs['v_out']

        outputs['Vdot'] = -v_out*A_out

    def compute_partials(self, inputs, partials):
        A_out = inputs['A_out']
        v_out = inputs['v_out']

        partials['Vdot', 'A_out'] = -v_out
        partials['Vdot', 'v_out'] = -A_out


class _PressureRate(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(name='p', val=np.ones(nn), desc='air pressure', units='N/m**2')
        self.add_input(name='k', val=1.4*np.ones(nn), desc='polytropic coefficient for expansion', units=None)
        self.add_input(name='V_b', val=2e-3*np.ones(nn), desc='bottle volume', units='m**3')
        self.add_input(name='V_w', val=1e-3*np.ones(nn), desc='water volume', units='m**3')
        self.add_input(name='Vdot', shape=(nn,), desc='water flow', units='m**3/s')

        self.add_output(name='pdot', shape=(nn,), desc='pressure derivative', units='N/m**2/s')

        ar = np.arange(nn)

        self.declare_partials(of='*', wrt='*', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        p = inputs['p']
        k = inputs['k']
        V_b = inputs['V_b']
        V_w = inputs['V_w']
        Vdot = inputs['Vdot']

        pdot = p*k*Vdot/(V_b-V_w)

        outputs['pdot'] = pdot

    def compute_partials(self, inputs, partials):
        p = inputs['p']
        k = inputs['k']
        V_b = inputs['V_b']
        V_w = inputs['V_w']
        Vdot = inputs['Vdot']

        partials['pdot', 'p'] = k*Vdot/(V_b-V_w)
        partials['pdot', 'k'] = p*Vdot/(V_b-V_w)
        partials['pdot', 'V_b'] = -p*k*Vdot/(V_b-V_w)**2
        partials['pdot', 'V_w'] = p*k*Vdot/(V_b-V_w)**2
        partials['pdot', 'Vdot'] = p*k/(V_b-V_w)


class _WaterThrust(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(name='rho_w', val=1e3*np.ones(nn), desc='water density', units='kg/m**3')
        self.add_input(name='A_out', val=np.pi*13e-3**2/4*np.ones(nn), desc='nozzle outlet area', units='m**2')
        self.add_input(name='v_out', val=np.zeros(nn), desc='water exhaust speed', units='m/s')

        self.add_output(name='F', shape=(nn,), desc='thrust', units='N')

        ar = np.arange(nn)

        self.declare_partials(of='*', wrt='*', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        rho_w = inputs['rho_w']
        A_out = inputs['A_out']
        v_out = inputs['v_out']

        outputs['F'] = rho_w*v_out**2*A_out

    def compute_partials(self, inputs, partials):
        rho_w = inputs['rho_w']
        A_out = inputs['A_out']
        v_out = inputs['v_out']

        partials['F', 'A_out'] = rho_w*v_out**2
        partials['F', 'rho_w'] = v_out**2*A_out
        partials['F', 'v_out'] = 2*rho_w*v_out*A_out


if __name__ == '__main__':
    p = om.Problem()
    p.model = WaterEngine(num_nodes=1)
    p.setup()
    p.final_setup()



import unittest

import numpy as np
import openmdao.api as om
import dymos as dm

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs
from dymos.utils.misc import om_version


class SimpleIVPSolution(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('t', shape=(nn,), units='s')
        self.add_input('y0', shape=(1,), units='unitless', tags=['dymos.static_target'])
        self.add_output('y', shape=(nn,), units='unitless')

        ar = np.arange(nn, dtype=int)
        self.declare_partials(of='y', wrt='t', rows=ar, cols=ar)
        self.declare_partials(of='y', wrt='y0', rows=ar, cols=np.zeros_like(ar))

    def compute(self, inputs, outputs):
        t = inputs['t']
        y0 = inputs['y0']
        outputs['y'] = t ** 2 + 2 * t + 1 - y0 * np.exp(t)

    def compute_partials(self, inputs, partials):
        t = inputs['t']
        y0 = inputs['y0']
        partials['y', 't'] = 2 * t + 2 - y0 * np.exp(t)
        partials['y', 'y0'] = -np.exp(t)


class SimpleBVPSolution(om.ExplicitComponent):
    """
    A basic BVP ODE solution taken from
    https://math.libretexts.org/Bookshelves/Applied_Mathematics/Numerical_Methods_(Chasnov)/07%3A_Ordinary_Differential_Equations/7.01%3A_Examples_of_Analytical_Solutions

    Find the solution to

    /begin{align}
        -\frac{d^{2} y}{d x^{2}}= x (1 - x)
    /end{align}

    subject to

    /begin{align}
        y(0) = 0 \\
        y(1) = 0
    /end{align}
    """

    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('x', shape=(nn,), units='s')
        self.add_input('y0', shape=(1,), units='unitless', tags=['dymos.static_target'])
        self.add_input('y1', shape=(1,), units='unitless', tags=['dymos.static_target'])
        self.add_output('y', shape=(nn,), units='unitless', tags=['dymos.state_source:y'])
        self.declare_coloring(method='cs', tol=1.0E-12)

    def compute(self, inputs, outputs):
        x = inputs['x']
        y_0 = inputs['y0']
        y_1 = inputs['y1']
        c_2 = y_0
        c_1 = y_1 - 1/12 + 1/6 - c_2
        outputs['y'] = x ** 4 / 12 - x ** 3 / 6 + c_1 * x + c_2


@use_tempdirs
class TestAnalyticPhaseSimpleResults(unittest.TestCase):

    def test_simple_ivp_system(self):

        p = om.Problem()
        p.model.add_subsystem('simple_ivp_solution', SimpleIVPSolution(num_nodes=100), promotes=['*'])
        p.setup(force_alloc_complex=True)

        p.set_val('t', np.linspace(0, 2, 100))
        p.set_val('y0', 0.5)

        p.run_model()

        t = p.get_val('t')
        y = p.get_val('y')

        expected_y = t ** 2 + 2 * t + 1 - 0.5 * np.exp(t)

        assert_near_equal(y, expected_y)

        p.check_partials(compact_print=True, method='cs', out_stream=None)

    def test_simple_ivp_system(self):

        p = om.Problem()
        p.model.add_subsystem('simple_ivp_solution', SimpleIVPSolution(num_nodes=100), promotes=['*'])
        p.setup(force_alloc_complex=True)

        p.set_val('t', np.linspace(0, 2, 100))
        p.set_val('y0', 0.5)

        p.run_model()

        t = p.get_val('t')
        y = p.get_val('y')

        expected_y = t ** 2 + 2 * t + 1 - 0.5 * np.exp(t)

        assert_near_equal(y, expected_y)

        p.check_partials(compact_print=True, method='cs', out_stream=None)

    def test_simple_ivp(self):

        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)

        traj.add_phase('phase', phase)

        phase.set_time_options(units='s', targets=['t'], fix_initial=True, fix_duration=True)
        phase.add_state('y')
        phase.add_parameter('y0', opt=False)

        p.setup()

        p.set_val('traj.phase.t_initial', 0.0, units='s')
        p.set_val('traj.phase.t_duration', 2.0, units='s')
        p.set_val('traj.phase.parameters:y0', 0.5, units='unitless')

        p.run_model()

        t = p.get_val('traj.phase.timeseries.time', units='s')
        y = p.get_val('traj.phase.timeseries.y', units='unitless')

        expected = t ** 2 + 2 * t + 1 - 0.5 * np.exp(t)

        assert_near_equal(y, expected)

    def test_simple_ivp_calc_expr(self):

        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)

        phase.add_calc_expr('y2 = y**2')

        traj.add_phase('phase', phase)

        phase.set_time_options(units='s', targets=['t'], fix_initial=True, fix_duration=True)
        phase.add_state('y')
        phase.add_parameter('y0', opt=False)

        p.setup()

        p.set_val('traj.phase.t_initial', 0.0, units='s')
        p.set_val('traj.phase.t_duration', 2.0, units='s')
        p.set_val('traj.phase.parameters:y0', 0.5, units='unitless')

        p.run_model()

        t = p.get_val('traj.phase.timeseries.time', units='s')
        y = p.get_val('traj.phase.timeseries.y', units='unitless')
        y2 = p.get_val('traj.phase.timeseries.y2')

        expected = t ** 2 + 2 * t + 1 - 0.5 * np.exp(t)

        assert_near_equal(y, expected)
        assert_near_equal(y2, y*y)

    def test_simple_bvp(self):

        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SimpleBVPSolution, num_nodes=11)

        traj.add_phase('phase', phase)

        phase.set_time_options(units='s', targets=['x'], fix_initial=True, fix_duration=True,
                               initial_val=0.0, duration_val=1.0)
        phase.add_parameter('y0', opt=False, val=0.0)
        phase.add_parameter('y1', opt=False, val=0.0)

        p.setup()

        p.run_model()

        t = p.get_val('traj.phase.timeseries.time', units='s')
        y = p.get_val('traj.phase.timeseries.y', units='unitless')

        expected = t * (1 - t) * (1 + t - t**2) / 12

        assert_near_equal(y, expected)

    def test_duplicate(self):

        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SimpleBVPSolution, num_nodes=11)

        phase.set_time_options(units='s', targets=['x'], fix_initial=True, fix_duration=True,
                               initial_val=0.0, duration_val=1.0)
        phase.add_parameter('y0', opt=False, val=0.0)
        phase.add_parameter('y1', opt=False, val=0.0)

        phase2 = phase.duplicate()
        traj.add_phase('phase2', phase2)

        p.setup()

        p.run_model()

        t = p.get_val('traj.phase2.timeseries.time', units='s')
        y = p.get_val('traj.phase2.timeseries.y', units='unitless')

        expected = t * (1 - t) * (1 + t - t**2) / 12

        assert_near_equal(y, expected)

    def test_renamed_state(self):

        class SolutionWithRenamedState(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=(int,))

            def setup(self):
                nn = self.options['num_nodes']
                self.add_input('t', shape=(nn,), units='s')
                self.add_input('y0', shape=(1,), units='unitless', tags=['dymos.static_target'])
                self.add_output('y', shape=(nn,), units='unitless', tags=['dymos.state_source:foo'])

                ar = np.arange(nn, dtype=int)
                self.declare_partials(of='y', wrt='t', rows=ar, cols=ar)
                self.declare_partials(of='y', wrt='y0', rows=ar, cols=np.zeros_like(ar))

            def compute(self, inputs, outputs):
                t = inputs['t']
                y0 = inputs['y0']
                outputs['y'] = t ** 2 + 2 * t + 1 - y0 * np.exp(t)

            def compute_partials(self, inputs, partials):
                t = inputs['t']
                y0 = inputs['y0']
                partials['y', 't'] = 2 * t + 2 - y0 * np.exp(t)
                partials['y', 'y0'] = -np.exp(t)

        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SolutionWithRenamedState, num_nodes=10)

        traj.add_phase('phase', phase)

        phase.set_time_options(units='s', targets=['t'], fix_initial=True, fix_duration=True)
        phase.add_parameter('y0', opt=False)

        p.setup()

        p.set_val('traj.phase.t_initial', 0.0, units='s')
        p.set_val('traj.phase.t_duration', 2.0, units='s')
        p.set_val('traj.phase.parameters:y0', 0.5, units='unitless')

        p.run_model()
        t = p.get_val('traj.phase.timeseries.time', units='s')
        y = p.get_val('traj.phase.timeseries.foo', units='unitless')

        expected = t ** 2 + 2 * t + 1 - 0.5 * np.exp(t)

        assert_near_equal(y, expected)

    def test_analytic_phase_load_case(self):

        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)

        traj.add_phase('phase', phase)

        phase.set_time_options(units='s', targets=['t'], fix_initial=True, fix_duration=True)
        phase.add_state('y')
        phase.add_parameter('y0', opt=False)

        p.setup()

        p.set_val('traj.phase.t_initial', 0.0, units='s')
        p.set_val('traj.phase.t_duration', 2.0, units='s')
        p.set_val('traj.phase.parameters:y0', 0.5, units='unitless')

        dm.run_problem(p, simulate=False)

        # Change the inputs so that we get different answers unless we reload the same case.
        p.set_val('traj.phase.t_initial', 0.0, units='s')
        p.set_val('traj.phase.t_duration', 1.0, units='s')
        p.set_val('traj.phase.parameters:y0', 0.6, units='unitless')

        sol_db = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db

        # Load the previous case and rerun
        dm.run_problem(p, simulate=False, restart=sol_db)

        t = p.get_val('traj.phase.timeseries.time', units='s')
        y = p.get_val('traj.phase.timeseries.y', units='unitless')

        expected = t ** 2 + 2 * t + 1 - 0.5 * np.exp(t)

        assert_near_equal(y, expected)


@use_tempdirs
class TestAnalyticPhaseInvalidOptions(unittest.TestCase):

    def test_add_control(self):
        phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=11)

        with self.assertRaises(NotImplementedError) as e:
            phase.add_control('foo')

        self.assertEqual('AnalyticPhase does not support controls.', str(e.exception))

    def test_set_control_options(self):
        phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=11)

        with self.assertRaises(NotImplementedError) as e:
            phase.set_control_options('foo', lower=0)

        self.assertEqual('AnalyticPhase does not support controls.', str(e.exception))

    def test_timeseries_expr(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)

        traj.add_phase('phase', phase)

        phase.set_time_options(units='s', targets=['t'], fix_initial=True, fix_duration=True)
        phase.add_state('y')
        phase.add_parameter('y0', units='unitless', opt=False)
        phase.add_timeseries_output('z = y0 + y**2', y0={'units': 'unitless'})

        p.setup()

        p.set_val('traj.phase.t_initial', 0.0, units='s')
        p.set_val('traj.phase.t_duration', 2.0, units='s')
        p.set_val('traj.phase.parameters:y0', 0.5, units='unitless')

        p.run_model()

        y = p.get_val('traj.phase.timeseries.y', units='unitless')
        y0 = p.get_val('traj.phase.parameter_vals:y0', units='unitless')

        expected_z = y0 + y**2
        z = p.get_val('traj.phase.timeseries.z')

        assert_near_equal(z, expected_z)


class TestLinkedAnalyticPhases(unittest.TestCase):

    def test_linked_phases_connected_false(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        first_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        first_phase.set_time_options(units='s', targets=['t'], fix_initial=True, duration_bounds=(0.5, 10.0))
        first_phase.add_state('y')
        first_phase.add_parameter('y0', opt=False)

        first_phase.add_boundary_constraint('y', loc='final', equals=1.5, units='unitless')

        second_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        second_phase.set_time_options(units='s', targets=['t'], fix_initial=False, duration_bounds=(0.1, 10.0))
        second_phase.add_state('y')
        second_phase.add_parameter('y0', opt=True)

        second_phase.add_boundary_constraint('time', loc='final', equals=2.0, units='s')

        # Since we're using constraints to enforce continuity between the two phases, we need a
        # driver and a dummy objective.  As usual, time works well for a dummy objective here.
        first_phase.add_objective('time', loc='final')
        p.driver = om.ScipyOptimizeDriver()

        traj.add_phase('first_phase', first_phase)
        traj.add_phase('second_phase', second_phase)

        traj.link_phases(['first_phase', 'second_phase'], ['time', 'y'], connected=False)

        p.setup()

        p.set_val('traj.first_phase.t_initial', 0.0, units='s')
        p.set_val('traj.first_phase.t_duration', 2.0, units='s')
        p.set_val('traj.first_phase.parameters:y0', 0.5, units='unitless')

        p.run_driver()

        t_1 = p.get_val('traj.first_phase.timeseries.time', units='s')
        x_1 = p.get_val('traj.first_phase.timeseries.y', units='unitless')
        y0_1 = p.get_val('traj.first_phase.parameter_vals:y0')

        t_2 = p.get_val('traj.second_phase.timeseries.time', units='s')
        x_2 = p.get_val('traj.second_phase.timeseries.y', units='unitless')
        y0_2 = p.get_val('traj.second_phase.parameter_vals:y0')

        assert_near_equal(1.500000, x_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(0.5338712554624387, t_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(2.0, t_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(5.305471950533106, x_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(y0_1, y0_2, tolerance=1.0E-6)

    def test_linked_phases_connected_time(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        first_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        first_phase.set_time_options(units='s', targets=['t'], fix_initial=True, duration_bounds=(0.5, 10.0))
        first_phase.add_state('y')
        first_phase.add_parameter('y0', opt=False)

        first_phase.add_boundary_constraint('y', loc='final', equals=1.5, units='unitless')

        second_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        second_phase.set_time_options(units='s', targets=['t'], fix_initial=False, duration_bounds=(0.1, 10.0))
        second_phase.add_state('y')
        second_phase.add_parameter('y0', opt=True)

        second_phase.add_boundary_constraint('time', loc='final', equals=2.0, units='s')

        # Since we're using constraints to enforce continuity between the two phases, we need a
        # driver and a dummy objective.  As usual, time works well for a dummy objective here.
        first_phase.add_objective('time', loc='final')
        p.driver = om.ScipyOptimizeDriver()

        traj.add_phase('first_phase', first_phase)
        traj.add_phase('second_phase', second_phase)

        # We can link time with a connection, since initial time is an input to the second phase.
        traj.link_phases(['first_phase', 'second_phase'], ['time'], connected=True)
        traj.link_phases(['first_phase', 'second_phase'], ['y'], connected=False)

        p.setup()

        p.set_val('traj.first_phase.t_initial', 0.0, units='s')
        p.set_val('traj.first_phase.t_duration', 2.0, units='s')
        p.set_val('traj.first_phase.parameters:y0', 0.5, units='unitless')

        p.run_driver()

        t_1 = p.get_val('traj.first_phase.timeseries.time', units='s')
        x_1 = p.get_val('traj.first_phase.timeseries.y', units='unitless')
        y0_1 = p.get_val('traj.first_phase.parameter_vals:y0')

        t_2 = p.get_val('traj.second_phase.timeseries.time', units='s')
        x_2 = p.get_val('traj.second_phase.timeseries.y', units='unitless')
        y0_2 = p.get_val('traj.second_phase.parameter_vals:y0')

        assert_near_equal(1.500000, x_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(0.5338712554624387, t_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(2.0, t_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(5.305471950533106, x_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(y0_1, y0_2, tolerance=1.0E-6)

    def test_linked_phases_connected_state(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        first_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        first_phase.set_time_options(units='s', targets=['t'], fix_initial=True, duration_bounds=(0.5, 10.0))
        first_phase.add_state('y')
        first_phase.add_parameter('y0', opt=False)

        first_phase.add_boundary_constraint('y', loc='final', equals=1.5, units='unitless')

        second_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        second_phase.set_time_options(units='s', targets=['t'], fix_initial=False, duration_bounds=(0.1, 10.0))
        second_phase.add_state('y')
        second_phase.add_parameter('y0', opt=True)

        second_phase.add_boundary_constraint('time', loc='final', equals=2.0, units='s')

        # Since we're using constraints to enforce continuity between the two phases, we need a
        # driver and a dummy objective.  As usual, time works well for a dummy objective here.
        first_phase.add_objective('time', loc='final')
        p.driver = om.ScipyOptimizeDriver()

        traj.add_phase('first_phase', first_phase)
        traj.add_phase('second_phase', second_phase)

        # We can link time with a connection, since initial time is an input to the second phase.
        traj.link_phases(['first_phase', 'second_phase'], ['time'], connected=True)
        traj.link_phases(['first_phase', 'second_phase'], ['y'], connected=True)

        with self.assertRaises(Exception) as e:
            p.setup()

        expected = "traj: Phase `first_phase` links variable `y` to phase `second_phase` state variable `y` by " \
                   "connection, but phase `second_phase` is an AnalyticPhase and does not support linking initial " \
                   "state values with option `connected=True`."

        self.assertEqual(expected, str(e.exception))

    def test_common_traj_param(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        first_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        first_phase.set_time_options(units='s', targets=['t'], fix_initial=True, duration_bounds=(0.5, 10.0))
        first_phase.add_state('y')
        first_phase.add_parameter('y0', opt=False)

        first_phase.add_boundary_constraint('y', loc='final', equals=1.5, units='unitless')

        second_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        second_phase.set_time_options(units='s', targets=['t'], fix_initial=False, duration_bounds=(0.1, 10.0))
        second_phase.add_state('y')
        second_phase.add_parameter('y0', opt=False)
        second_phase.add_boundary_constraint('time', loc='final', equals=2.0, units='s')

        # Since we're using constraints to enforce continuity between the two phases, we need a
        # driver and a dummy objective.  As usual, time works well for a dummy objective here.
        first_phase.add_objective('time', loc='final')
        p.driver = om.ScipyOptimizeDriver()

        traj.add_phase('first_phase', first_phase)
        traj.add_phase('second_phase', second_phase)

        # We can link time with a connection, since initial time is an input to the second phase.
        traj.link_phases(['first_phase', 'second_phase'], ['time'], connected=True)

        # Make the y0 parameter common between the two phases, since the particular solution to the ODE applies to both
        traj.add_parameter('y0', opt=False)

        p.setup()

        p.set_val('traj.first_phase.t_initial', 0.0, units='s')
        p.set_val('traj.first_phase.t_duration', 2.0, units='s')
        p.set_val('traj.parameters:y0', 0.5, units='unitless')

        p.run_driver()

        t_1 = p.get_val('traj.first_phase.timeseries.time', units='s')
        x_1 = p.get_val('traj.first_phase.timeseries.y', units='unitless')
        y0_1 = p.get_val('traj.first_phase.parameter_vals:y0')

        t_2 = p.get_val('traj.second_phase.timeseries.time', units='s')
        x_2 = p.get_val('traj.second_phase.timeseries.y', units='unitless')
        y0_2 = p.get_val('traj.second_phase.parameter_vals:y0')

        assert_near_equal(1.500000, x_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(0.5338712554624387, t_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(2.0, t_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(5.305471950533106, x_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(y0_1, y0_2, tolerance=1.0E-6)

    def test_link_params_connected_false(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        first_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        first_phase.set_time_options(units='s', targets=['t'], fix_initial=True, duration_bounds=(0.5, 10.0))
        first_phase.add_state('y')
        first_phase.add_parameter('y0', opt=False)

        first_phase.add_boundary_constraint('y', loc='final', equals=1.5, units='unitless')

        second_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        second_phase.set_time_options(units='s', targets=['t'], fix_initial=False, duration_bounds=(0.1, 10.0))
        second_phase.add_state('y')
        second_phase.add_parameter('y0', opt=True)
        second_phase.add_boundary_constraint('time', loc='final', equals=2.0, units='s')

        # Since we're using constraints to enforce continuity between the two phases, we need a
        # driver and a dummy objective.  As usual, time works well for a dummy objective here.
        first_phase.add_objective('time', loc='final')
        p.driver = om.ScipyOptimizeDriver()

        traj.add_phase('first_phase', first_phase)
        traj.add_phase('second_phase', second_phase)

        # We can link time with a connection, since initial time is an input to the second phase.
        traj.link_phases(['first_phase', 'second_phase'], ['time'], connected=True)
        traj.link_phases(['first_phase', 'second_phase'], ['y0'], connected=False)

        p.setup()

        p.set_val('traj.first_phase.t_initial', 0.0, units='s')
        p.set_val('traj.first_phase.t_duration', 2.0, units='s')
        p.set_val('traj.first_phase.parameters:y0', 0.5, units='unitless')

        p.run_driver()

        t_1 = p.get_val('traj.first_phase.timeseries.time', units='s')
        x_1 = p.get_val('traj.first_phase.timeseries.y', units='unitless')
        y0_1 = p.get_val('traj.first_phase.parameter_vals:y0')

        t_2 = p.get_val('traj.second_phase.timeseries.time', units='s')
        x_2 = p.get_val('traj.second_phase.timeseries.y', units='unitless')
        y0_2 = p.get_val('traj.second_phase.parameter_vals:y0')

        assert_near_equal(1.500000, x_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(0.5338712554624387, t_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(2.0, t_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(5.305471950533106, x_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(y0_1, y0_2, tolerance=1.0E-6)

    def test_link_params_connected_true(self):
        p = om.Problem()
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        first_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        first_phase.set_time_options(units='s', targets=['t'], fix_initial=True, duration_bounds=(0.5, 10.0))
        first_phase.add_state('y')
        first_phase.add_parameter('y0', opt=False)

        first_phase.add_boundary_constraint('y', loc='final', equals=1.5, units='unitless')

        second_phase = dm.AnalyticPhase(ode_class=SimpleIVPSolution, num_nodes=10)
        second_phase.set_time_options(units='s', targets=['t'], fix_initial=False, duration_bounds=(0.1, 10.0))
        second_phase.add_state('y')
        second_phase.add_parameter('y0', opt=True)
        second_phase.add_boundary_constraint('time', loc='final', equals=2.0, units='s')

        # Since we're using constraints to enforce continuity between the two phases, we need a
        # driver and a dummy objective.  As usual, time works well for a dummy objective here.
        first_phase.add_objective('time', loc='final')
        p.driver = om.ScipyOptimizeDriver()

        traj.add_phase('first_phase', first_phase)
        traj.add_phase('second_phase', second_phase)

        # We can link time with a connection, since initial time is an input to the second phase.
        traj.link_phases(['first_phase', 'second_phase'], ['time'], connected=True)
        traj.link_phases(['first_phase', 'second_phase'], ['y0'], connected=True)

        p.setup()

        p.set_val('traj.first_phase.t_initial', 0.0, units='s')
        p.set_val('traj.first_phase.t_duration', 2.0, units='s')
        p.set_val('traj.first_phase.parameters:y0', 0.5, units='unitless')

        dm.run_problem(p, simulate=False)

        t_1 = p.get_val('traj.first_phase.timeseries.time', units='s')
        x_1 = p.get_val('traj.first_phase.timeseries.y', units='unitless')

        t_2 = p.get_val('traj.second_phase.timeseries.time', units='s')
        x_2 = p.get_val('traj.second_phase.timeseries.y', units='unitless')

        assert_near_equal(1.500000, x_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(0.5338712554624387, t_1[-1, 0], tolerance=1.0E-6)
        assert_near_equal(2.0, t_2[-1, 0], tolerance=1.0E-6)
        assert_near_equal(5.305471950533106, x_2[-1, 0], tolerance=1.0E-6)

        with self.assertRaises(RuntimeError) as e:
            dm.run_problem(p, simulate=True)

        expected = 'Trajectory `traj` has no phases that support simulation.'
        self.assertEqual(expected, str(e.exception))



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs


# Modify class so we can run it standalone.
import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.examples.brachistochrone.test.ex_brachistochrone import brachistochrone_min_time as brach


@use_tempdirs
class TestCollocationBalanceIndex(unittest.TestCase):
    """
    Test that the indices used in the StateIndependentsComp are as expected.
    """

    def setUp(self):
        dm.options['include_check_partials'] = True

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_3_lgl(self):
        """
        Test one 3rd order LGL segment indices

        All nodes are:   [0, 1, 2]
        Input nodes:     [^     ^]
        Solver nodes:    [      ^]
        Indep nodes:     [^      ]
        Of input nodes, solver nodes for fix_initial are: {1}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='gauss-lobatto', num_segments=1, transcription_order=3,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_5_lgl(self):
        """
        Test one 5th order LGL segment indices

        All nodes are:   [0, 1, 2, 3, 4]
        Input nodes:     [^     ^     ^]
        Solver nodes:    [      ^     ^]
        Of input nodes, solver nodes for fix_initial are: {1, 2}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='gauss-lobatto', num_segments=1, transcription_order=5,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_3_lgl_compressed(self):
        """
        Test one 3rd order LGL segment indices

        All nodes are:   [0, 1, 2, 3, 4]
        Input nodes:     [^     ^     ^]
        Solver nodes:    [      ^     ^]
        Indep nodes:     [^            ]
        Of input nodes, solver nodes for fix_initial are: {1, 2}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='gauss-lobatto', num_segments=2, transcription_order=3,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_5_lgl_compressed(self):
        """
        Test two 5th order LGL segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        Input nodes:     [^     ^     ^        ^     ^]
        Solver nodes:    [      ^     ^        ^     ^]
        Indep nodes:     [^                        ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3, 4}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='gauss-lobatto', num_segments=2, transcription_order=5,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3, 4})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_3_lgl_uncompressed(self):
        """
        Test two 3rd order LGL segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5]
        Input nodes:     [^     ^  ^     ^]
        Solver nodes:    [      ^        ^]
        Indep nodes:     [^        ^      ]
        Of input nodes, solver nodes for fix_initial are: {1, 3}
        The indep node is just the first one: {0, 2}
        """
        p = brach(transcription='gauss-lobatto', num_segments=2, transcription_order=3,
                  compressed=False, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 3})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0, 2})

    def test_5_lgl_uncompressed(self):
        """
        Test two 5th order LGL segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        Input nodes:     [^     ^     ^  ^     ^     ^]
        Solver nodes:    [      ^     ^        ^     ^]
        Indep nodes:     [^              ^         ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 4, 5}
        The indep node is just the first one in each segment: {0, 3}
        """
        p = brach(transcription='gauss-lobatto', num_segments=2, transcription_order=5,
                  compressed=False, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 4, 5})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0, 3})

    def test_3_radau(self):
        """
        Test one 3rd order radau segment indices

        All nodes are:   [0, 1, 2, 3]
        Input nodes:     [^  ^  ^  ^]
        Solver nodes:    [   ^  ^  ^]
        Indep nodes:     [^         ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='radau-ps', num_segments=1, transcription_order=3,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_5_radau(self):
        """
        Test one 5th order radau segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5]
        Input nodes:     [^  ^  ^  ^  ^  ^]
        Solver nodes:    [   ^  ^  ^  ^  ^]
        Indep nodes:     [^         ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3, 4, 5}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='radau-ps', num_segments=1, transcription_order=5,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3, 4, 5})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_3_radau_compressed(self):
        """
        Test one 3rd order radau segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5, 6, 7]
        Input nodes:     [^  ^  ^  ^     ^  ^  ^]
        Solver nodes:    [   ^  ^  ^     ^  ^  ^]
        Indep nodes:     [^            ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3, 4, 5, 6}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='radau-ps', num_segments=2, transcription_order=3,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3, 4, 5, 6})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_5_radau_compressed(self):
        """
        Test two 5th order radau segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        Input nodes:     [^  ^  ^  ^  ^  ^     ^  ^  ^   ^   ^]
        Solver nodes:    [   ^  ^  ^  ^  ^     ^  ^  ^   ^   ^]
        Indep nodes:     [^                                   ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}
        The indep node is just the first one: {0}
        """
        p = brach(transcription='radau-ps', num_segments=2, transcription_order=5,
                  compressed=True, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3, 4, 5, 6, 7, 8, 9, 10})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0})

    def test_3_radau_uncompressed(self):
        """
        Test one 3rd order radau segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5, 6, 7]
        Input nodes:     [^  ^  ^  ^  ^  ^  ^  ^]
        Solver nodes:    [   ^  ^  ^     ^  ^  ^]
        Indep nodes:     [^           ^         ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3, 5, 6, 7}
        The indep node is just the first one: {0, 4}
        """
        p = brach(transcription='radau-ps', num_segments=2, transcription_order=3,
                  compressed=False, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3, 5, 6, 7})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0, 4})

    def test_5_radau_uncompressed(self):
        """
        Test two 5th order radau segment indices

        All nodes are:   [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        Input nodes:     [^  ^  ^  ^  ^  ^  ^  ^  ^  ^   ^   ^]
        Solver nodes:    [   ^  ^  ^  ^  ^     ^  ^  ^   ^   ^]
        Indep nodes:     [^                 ^                 ]
        Of input nodes, solver nodes for fix_initial are: {1, 2, 3, 4, 5, 7, 8, 9, 10, 11}
        The indep node is just the first one: {0, 6}
        """
        p = brach(transcription='radau-ps', num_segments=2, transcription_order=5,
                  compressed=False, solve_segments='forward', run_driver=True)

        state_indeps_comp = p.model.traj0.phases.phase0.indep_states

        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['solver']), {1, 2, 3, 4, 5, 7, 8, 9, 10, 11})
        self.assertSetEqual(set(state_indeps_comp.state_idx_map['x']['indep']), {0, 6})


@use_tempdirs
class TestCollocationBalanceApplyNL(unittest.TestCase):

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def make_prob(self, transcription, num_segments, transcription_order, compressed):

        p = om.Problem(model=om.Group())

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=num_segments,
                                order=transcription_order,
                                compressed=compressed,)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=num_segments,
                         order=transcription_order,
                         compressed=compressed)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, solve_segments='forward')
        phase.add_state('y', fix_initial=True, fix_final=False, solve_segments='forward')

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments='forward')

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.setup(force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=2.0)
        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        return p

    def test_apply_nonlinear_gl(self):
        dm.options['include_check_partials'] = True
        p = self.make_prob(transcription='gauss-lobatto', num_segments=3, transcription_order=3,
                           compressed=True)

        p.final_setup()
        p.model.run_apply_nonlinear()  # need to make sure residuals are computed

        expected = np.array([[0., 1., 1., 1.]]).T

        outputs = p.model.traj0.phases.phase0.indep_states.list_outputs(residuals=True, out_stream=None)
        resids = {k: v['resids'] for k, v in outputs}

        assert_almost_equal(resids['states:x'], expected)
        assert_almost_equal(resids['states:v'], expected)

    def test_apply_nonlinear_radau(self):
        dm.options['include_check_partials'] = True
        p = self.make_prob(transcription='radau-ps', num_segments=3, transcription_order=3,
                           compressed=True)

        p.final_setup()
        p.model.run_apply_nonlinear()  # need to make sure residuals are computed

        expected = np.array([[0., 1., 1., 1., 1., 1., 1., 1., 1., 1.]]).T

        outputs = p.model.traj0.phases.phase0.indep_states.list_outputs(residuals=True, out_stream=None)
        resids = {k: v['resids'] for k, v in outputs}

        assert_almost_equal(resids['states:x'], expected)
        assert_almost_equal(resids['states:v'], expected)

    def test_partials_gl(self):
        dm.options['include_check_partials'] = True
        p = self.make_prob(transcription='gauss-lobatto', num_segments=3, transcription_order=3,
                           compressed=True)

        def assert_partials(data):
            for of, wrt in data:
                if of == wrt:
                    # IndepVarComp like outputs have correct derivs, but FD is wrong so we skip
                    # them (should be some form of -I)
                    continue
                check_data = data[(of, wrt)]
                fwd_abs_error = check_data['abs error'].forward
                if fwd_abs_error is not None:
                    self.assertLess(check_data['abs error'].forward, 1e-8)

        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)
        data = cpd['traj0.phases.phase0.indep_states']
        assert_partials(data)

    def test_partials_radau(self):
        dm.options['include_check_partials'] = True
        p = self.make_prob(transcription='radau-ps', num_segments=3, transcription_order=3,
                           compressed=True)

        def assert_partials(data):
            for of, wrt in data:
                if of == wrt:
                    # IndepVarComp like outputs have correct derivs, but FD is wrong so we skip
                    # them (should be some form of -I)
                    continue
                check_data = data[(of, wrt)]
                fwd_abs_error = check_data['abs error'].forward
                if fwd_abs_error is not None:
                    self.assertLess(check_data['abs error'].forward, 1e-8)

        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)
        data = cpd['traj0.phases.phase0.indep_states']
        assert_partials(data)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import openmdao.api as om
import numpy as np


class BreakwellODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('u', val=np.zeros(nn))
        self.add_output('J_dot', val=np.zeros(nn))

        ar = np.arange(nn)
        self.declare_partials(of='J_dot', wrt='u', rows=ar, cols=ar)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        u = inputs['u']
        outputs['J_dot'] = 0.5*u**2

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        u = inputs['u']
        partials['J_dot', 'u'] = u



import numpy as np
import openmdao.api as om
import dymos as dm
import unittest

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from dymos.examples.moon_landing import MoonLandingProblemODE


@use_tempdirs
class TestMoonLandingProblem(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def make_problem(self, grid_type):
        self.p = om.Problem(model=om.Group())
        self.p.driver = om.pyOptSparseDriver()
        self.p.driver.declare_coloring()
        self.p.driver.options['optimizer'] = 'IPOPT'
        self.p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
        self.p.driver.opt_settings['print_level'] = 0
        self.p.driver.opt_settings['linear_solver'] = 'mumps'
        self.p.driver.declare_coloring()

        t = dm.Birkhoff(num_nodes=20, grid_type=grid_type)

        traj = self.p.model.add_subsystem('traj', dm.Trajectory())
        phase = dm.Phase(ode_class=MoonLandingProblemODE,
                         transcription=t)

        phase.set_time_options(fix_initial=True, fix_duration=False)
        phase.add_state('h', fix_initial=True, rate_source='h_dot')
        phase.add_state('v', fix_initial=True, rate_source='v_dot')
        phase.add_state('m', fix_initial=True, lower=1e-3, rate_source='m_dot')
        phase.add_control('T', lower=0.0, upper=1.227)

        phase.add_boundary_constraint('h', loc='final', equals=0.0)
        phase.add_boundary_constraint('v', loc='final', equals=0.0)

        phase.add_objective('m', scaler=-1)
        phase.set_simulate_options(atol=1.0E-1, rtol=1.0E-2)

        traj.add_phase('phase', phase)

        self.p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.0)
        phase.set_state_val('h', [1.0, 0.0])
        phase.set_state_val('v', [-0.783, 0.0])
        phase.set_state_val('m', [1.0, 0.2])
        phase.set_control_val('T', [0.0, 1.227])

    def test_problem_lgl(self):

        with dm.options.temporary(include_check_partials=True):
            self.make_problem(grid_type='lgl')
            dm.run_problem(self.p, simulate=True, simulate_kwargs={'times_per_seg': 100},
                           make_plots=True)

            with np.printoptions(linewidth=1024):
                self.p.check_partials(compact_print=True, method='cs')

        h = self.p.get_val('traj.phase.timeseries.h')
        v = self.p.get_val('traj.phase.timeseries.v')
        m = self.p.get_val('traj.phase.timeseries.m')
        T = self.p.get_val('traj.phase.timeseries.T')

        assert_near_equal(T[0], 0.0, tolerance=1e-5)
        assert_near_equal(T[-1], 1.227, tolerance=1e-5)
        assert_near_equal(h[-1], 0.0, tolerance=1e-5)
        assert_near_equal(v[-1], 0.0, tolerance=1e-5)
        assert_near_equal(m[-1], 0.3953, tolerance=1e-3)

    def test_problem_cgl(self):
        self.make_problem(grid_type='cgl')
        dm.run_problem(self.p, make_plots=True)
        h = self.p.get_val('traj.phase.timeseries.h')
        v = self.p.get_val('traj.phase.timeseries.v')
        m = self.p.get_val('traj.phase.timeseries.m')
        T = self.p.get_val('traj.phase.timeseries.T')

        assert_near_equal(T[0], 0.0, tolerance=1e-5)
        assert_near_equal(T[-1], 1.227, tolerance=1e-5)
        assert_near_equal(h[-1], 0.0, tolerance=1e-5)
        assert_near_equal(v[-1], 0.0, tolerance=1e-5)
        assert_near_equal(m[-1], 0.3953, tolerance=1e-4)



import time
import unittest

import numpy as np

import openmdao.api as om

import dymos
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.misc import GroupWrapperConfig, om_version
from dymos.utils.testing_utils import PhaseStub, SimpleODE
from dymos.transcriptions.picard_shooting.birkhoff_picard_iter_group import BirkhoffPicardIterGroup
from dymos.phase.options import StateOptionsDictionary, TimeOptionsDictionary
from dymos.transcriptions.grid_data import BirkhoffGrid, GaussLobattoGrid


BirkhoffPicardIterGroup = GroupWrapperConfig(BirkhoffPicardIterGroup, [PhaseStub()])


@use_tempdirs
@unittest.skipIf(om_version()[0] < (3, 37, 0), 'Requires OpenMDAO version later than 3.37.0')
class TestBirkhoffPicardIterGroup(unittest.TestCase):

    def test_birkhoff_picard_solve_segments(self):
        for direction in ['forward', 'backward']:
            for grid_type in ['lgl', 'cgl']:
                for nl_solver in ['newton', 'nlbgs']:
                    with self.subTest(msg=f'{direction=} {grid_type=} {nl_solver=}'):
                        with dymos.options.temporary(include_check_partials=True):

                            state_options = {'x': StateOptionsDictionary()}

                            state_options['x']['shape'] = (1,)
                            state_options['x']['units'] = 's**2'
                            state_options['x']['targets'] = ['x']

                            state_options['x']['initial_bounds'] = (None, None)
                            state_options['x']['final_bounds'] = (None, None)
                            state_options['x']['solve_segments'] = direction
                            state_options['x']['rate_source'] = 'x_dot'

                            time_options = TimeOptionsDictionary()
                            grid_data = GaussLobattoGrid(nodes_per_seg=11, num_segments=1)

                            nn = grid_data.subset_num_nodes['all']
                            ode_class = SimpleODE

                            p = om.Problem()
                            p.model.add_subsystem('birkhoff', BirkhoffPicardIterGroup(state_options=state_options,
                                                                                      time_units=time_options['units'],
                                                                                      grid_data=grid_data,
                                                                                      ode_class=ode_class))

                            birkhoff = p.model._get_subsystem('birkhoff')

                            if nl_solver == 'newton':
                                birkhoff.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, maxiter=100)
                            elif nl_solver == 'nlbgs':
                                birkhoff.nonlinear_solver = om.NonlinearBlockGS(maxiter=201, use_aitken=True)
                            else:
                                birkhoff.nonlinear_solver = om.NonlinearRunOnce()
                            birkhoff.linear_solver = om.DirectSolver()

                            p.setup(force_alloc_complex=True)

                            # Instead of using the TimeComp just transform the node segment taus onto [0, 2]
                            times = (grid_data.node_stau + 1) * 1.0

                            solution = np.reshape(times**2 + 2 * times + 1 - 0.5 * np.exp(times), (nn, 1))
                            dsolution_dt = np.reshape(2 * times + 2 - 0.5 * np.exp(times), (nn, 1))

                            p.final_setup()

                            if direction == 'forward':
                                p.set_val('birkhoff.picard_update_comp.seg_initial_states:x', 0.5)
                            else:
                                p.set_val('birkhoff.picard_update_comp.seg_final_states:x', solution[-1])
                            p.set_val('birkhoff.ode_all.t', times)
                            p.set_val('birkhoff.ode_all.p', 1.0)

                            p.final_setup()

                            t_start = time.perf_counter()
                            p.run_model()
                            t_end = time.perf_counter()

                            print(f"Elapsed time: {t_end-t_start:.4f} seconds")

                            assert_near_equal(solution, p.get_val('birkhoff.states:x'), tolerance=1.0E-9)
                            assert_near_equal(dsolution_dt.ravel(),
                                              p.get_val('birkhoff.picard_update_comp.f_computed:x').ravel(),
                                              tolerance=1.0E-9)

                            cpd = p.check_partials(method='cs', compact_print=True,
                                                   show_only_incorrect=True, out_stream=None)
                            assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @unittest.skip('This test is a demonstation of the inability of Birkhoff-Picard '
                   'iteration to solve highly nonlinear systems.')
    def test_birkhoff_solve_segments_lorenz(self):

        class LorenzAttractorODE(om.JaxExplicitComponent):
            """
            The ODE for the Lorenz attractor.
            """

            def initialize(self):
                """
                All Dymos ODE systems are required to have an option "num_nodes",
                which is the number of points at which the ODE is simultaneously evaluated.

                This will be set by the Phase during setup once the transcription details are known.
                """
                self.options.declare('num_nodes', types=(int,))

            def setup(self):
                """
                In setup, we add inputs and outputs.

                The first dimension is assumed to pertain to the index of the node.

                An input that's a scalar at each node should have a shape of
                (num_nodes, 1) or just (num_nodes,).

                For vectors or matrices, it's just the shape of the matrix at each
                node prepended with num_nodes.

                We provide units for the scalars, but OpenMDAO doesn't do unit conversion on an index-by-index basis,
                so we just assume that no unit conversion should be done for the S matrix and K vector.
                """
                nn = self.options['num_nodes']

                # ODE inputs
                self.add_input('x', shape=(nn,))
                self.add_input('y', shape=(nn,))
                self.add_input('z', shape=(nn,))
                self.add_input('s', shape=(1,))
                self.add_input('r', shape=(1,))
                self.add_input('b', shape=(1,))

                # State rates
                self.add_output('x_dot', shape=(nn,), units='1/s', tags=['dymos.state_rate_source:x'])
                self.add_output('y_dot', shape=(nn,), units='1/s', tags=['dymos.state_rate_source:y'])
                self.add_output('z_dot', shape=(nn,), units='1/s', tags=['dymos.state_rate_source:z'])

            # because our compute primal output depends on static variables, in this case
            # and self.options['num_nodes'], we must define a get_self_statics method. This method must
            # return a tuple of all static variables. Their order in the tuple doesn't matter.  If your
            # component happens to have discrete inputs, do NOT return them here. Discrete inputs are passed
            # into the compute_primal function individually, after the continuous variables.
            def get_self_statics(self):
                # return value must be hashable
                return self.options['num_nodes'],

            def setup_partials(self):
                nn = self.options['num_nodes']
                ar = np.arange(nn, dtype=int)
                self.declare_partials('x_dot', 'x', rows=ar, cols=ar)
                self.declare_partials('x_dot', 'y', rows=ar, cols=ar)
                self.declare_partials('x_dot', 's')

                self.declare_partials('y_dot', 'x', rows=ar, cols=ar)
                self.declare_partials('y_dot', 'y', rows=ar, cols=ar)
                self.declare_partials('y_dot', 'z', rows=ar, cols=ar)
                self.declare_partials('y_dot', 'r')

                self.declare_partials('z_dot', 'x', rows=ar, cols=ar)
                self.declare_partials('z_dot', 'y', rows=ar, cols=ar)
                self.declare_partials('z_dot', 'z', rows=ar, cols=ar)
                self.declare_partials('z_dot', 'b')

            def compute_primal(self, x, y, z, s=10, r=28, b=2.667):
                """
                Parameters
                ----------
                x : array-like
                    The x state.
                y : array-like
                    The y state.
                z : array-like
                    The z state.
                s : float
                    The Lorenz s parameter.
                r : float
                    The Lorenz r parameter.
                b : float
                    The Lorenz b parameter.

                Returns
                -------
                x_dot : array-like
                    The time-derivative of the x state.
                y_dot : array-like
                    The time-derivative of the y state.
                z_dot : array-like
                    The time-derivative of the z state.
                """
                x_dot = s * (y - x)
                y_dot = r * x - y - x*z
                z_dot = x * y - b * z
                return x_dot, y_dot, z_dot

        for direction in ['forward']:
            for grid_type in ['lgl']:
                with self.subTest(msg=grid_type):
                    with dymos.options.temporary(include_check_partials=True):

                        state_options = {'x': StateOptionsDictionary(),
                                         'y': StateOptionsDictionary(),
                                         'z': StateOptionsDictionary()}

                        state_options['x']['shape'] = (1,)
                        state_options['x']['units'] = None
                        state_options['x']['targets'] = ['x']
                        state_options['x']['initial_bounds'] = (None, None)
                        state_options['x']['final_bounds'] = (None, None)
                        state_options['x']['solve_segments'] = direction
                        state_options['x']['rate_source'] = 'x_dot'

                        state_options['y']['shape'] = (1,)
                        state_options['y']['units'] = None
                        state_options['y']['targets'] = ['y']
                        state_options['y']['initial_bounds'] = (None, None)
                        state_options['y']['final_bounds'] = (None, None)
                        state_options['y']['solve_segments'] = direction
                        state_options['y']['rate_source'] = 'y_dot'

                        state_options['z']['shape'] = (1,)
                        state_options['z']['units'] = None
                        state_options['z']['targets'] = ['z']
                        state_options['z']['initial_bounds'] = (None, None)
                        state_options['z']['final_bounds'] = (None, None)
                        state_options['z']['solve_segments'] = direction
                        state_options['z']['rate_source'] = 'z_dot'

                        time_options = TimeOptionsDictionary()
                        grid_data = BirkhoffGrid(num_nodes=300, grid_type=grid_type)
                        nn = grid_data.subset_num_nodes['all']
                        ode_class = LorenzAttractorODE

                        p = om.Problem()
                        p.model.add_subsystem('birkhoff', BirkhoffPicardIterGroup(state_options=state_options,
                                                                                  time_options=time_options,
                                                                                  grid_data=grid_data,
                                                                                  ode_class=ode_class))

                        birkhoff = p.model._get_subsystem('birkhoff')

                        birkhoff.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, maxiter=100, iprint=0)
                        birkhoff.nonlinear_solver = om.NonlinearBlockGS(maxiter=500, use_aitken=True, iprint=0)
                        birkhoff.linear_solver = om.DirectSolver()

                        p.setup(force_alloc_complex=True)

                        # Instead of using the TimeComp just transform the node segment taus onto [0, 100]
                        times = (grid_data.node_stau + 1) * 0.3
                        dt_dstau = (times[-1] / 2.0) * np.ones(nn)

                        if direction == 'forward':
                            p.set_val('birkhoff.initial_states:x', 0.0)
                            p.set_val('birkhoff.initial_states:y', 1.0)
                            p.set_val('birkhoff.initial_states:z', 1.05)

                            p.set_val('birkhoff.states:x', 0.0)
                            p.set_val('birkhoff.states:y', 1.0)
                            p.set_val('birkhoff.states:z', 1.05)
                        else:
                            raise ValueError('not set up for backwards prop yet')
                            # p.set_val('birkhoff.final_states:x', solution[-1])
                        p.set_val('birkhoff.dt_dstau', dt_dstau)
                        p.set_val('birkhoff.ode_all.s', 10.0)
                        p.set_val('birkhoff.ode_all.r', 28.0)
                        p.set_val('birkhoff.ode_all.b', 2.667)

                        p.final_setup()
                        p.run_model()

                    p.model.list_vars()

                    x = p.get_val('birkhoff.states:x')
                    y = p.get_val('birkhoff.states:y')
                    z = p.get_val('birkhoff.states:z')

                    import matplotlib.pyplot as plt
                    ax = plt.figure().add_subplot(projection='3d')
                    ax.plot(x, y, z, '-')
                    plt.show()

                    # assert_near_equal(solution, p.get_val('birkhoff.states:x'), tolerance=1.0E-9)
                    # assert_near_equal(dsolution_dt.ravel(), p.get_val('birkhoff.state_rates:x'), tolerance=1.0E-9)
                    # assert_near_equal(solution[np.newaxis, 0], p.get_val('birkhoff.initial_states:x'), tolerance=1.0E-9)
                    # assert_near_equal(solution[np.newaxis, -1], p.get_val('birkhoff.final_states:x'), tolerance=1.0E-9)

                    # cpd = p.check_partials(method='fd', compact_print=True, out_stream=None)
                    # assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)


if __name__ == '__main__':
    unittest.main()



import numpy as np

import openmdao.api as om


class CannonballSizeComp(om.ExplicitComponent):
    """
    Compute the reference area and mass of a cannonball with a given radius and density.

    Notes
    -----
    This component is not vectorized with 'num_nodes' as is the usual way with Dymos, but is instead
    intended to compute a scalar mass and reference area from scalar radius and density inputs. This
    component does not reside in the ODE but instead its outputs are connected to the trajectory via
    input design parameters.
    """
    def setup(self):
        self.add_input(name='radius', val=1.0, desc='cannonball radius', units='m')
        self.add_input(name='dens', val=7870., desc='cannonball density', units='kg/m**3')

        self.add_output(name='mass', shape=(1,), desc='cannonball mass', units='kg')
        self.add_output(name='S', shape=(1,), desc='aerodynamic reference area', units='m**2')

        self.declare_partials(of='mass', wrt='dens')
        self.declare_partials(of='mass', wrt='radius')

        self.declare_partials(of='S', wrt='radius')

    def compute(self, inputs, outputs):
        radius = inputs['radius']
        dens = inputs['dens']

        outputs['mass'] = (4/3.) * dens * np.pi * radius ** 3
        outputs['S'] = np.pi * radius ** 2

    def compute_partials(self, inputs, partials):
        radius = inputs['radius']
        dens = inputs['dens']

        partials['mass', 'dens'] = (4/3.) * np.pi * radius ** 3
        partials['mass', 'radius'] = 4. * dens * np.pi * radius ** 2

        partials['S', 'radius'] = 2 * np.pi * radius



import importlib
import os
import unittest
import numpy as np
from numpy.polynomial import Polynomial as P
try:
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
except ImportError:
    matpoltlib = None

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from dymos.utils.testing_utils import assert_timeseries_near_equal, _get_reports_dir
from dymos.utils.introspection import get_promoted_vars

import dymos as dm
from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE
from dymos.utils.misc import om_version
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse, set_env_vars_context


bokeh_available = importlib.util.find_spec('bokeh') is not None


def min_time_climb(optimizer='SLSQP', num_seg=3, transcription='gauss-lobatto',
                   transcription_order=3, force_alloc_complex=False, add_rate=False, time_name='time',
                   simulate=True, path_constraints=True, make_plots=False):

    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()

    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 1000
        p.driver.opt_settings['iSumm'] = 6
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        p.driver.opt_settings['Function precision'] = 1.0E-12
        p.driver.opt_settings['Linesearch tolerance'] = 0.1
        p.driver.opt_settings['Major step limit'] = 0.5
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['tol'] = 1.0E-5
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['mu_strategy'] = 'monotone'
        p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
        p.driver.opt_settings['mu_init'] = 0.01

    if transcription == 'gauss-lobatto':
        tx = dm.GaussLobatto(num_segments=num_seg, order=transcription_order)
    elif transcription == 'radau-ps':
        tx = dm.Radau(num_segments=num_seg, order=transcription_order)
    elif transcription == 'birkhoff':
        tx = dm.Birkhoff(num_nodes=transcription_order)

    traj = dm.Trajectory()

    phase = dm.Phase(ode_class=MinTimeClimbODE, transcription=tx)
    traj.add_phase('phase0', phase)

    p.model.add_subsystem('traj', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                           duration_ref=100.0, name=time_name)

    phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                    ref=1.0E3, defect_ref=1.0E3, units='m',
                    rate_source='flight_dynamics.r_dot')

    phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                    ref=20_000, defect_ref=20_000, units='m',
                    rate_source='flight_dynamics.h_dot', targets=['h'])

    phase.add_state('v', fix_initial=True, lower=10.0,
                    ref=1.0E2, defect_ref=1.0E2, units='m/s',
                    rate_source='flight_dynamics.v_dot', targets=['v'])

    phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                    ref=1.0, defect_ref=1.0, units='rad',
                    rate_source='flight_dynamics.gam_dot', targets=['gam'])

    phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                    ref=10_000, defect_ref=10_000, units='kg',
                    rate_source='prop.m_dot', targets=['m'])

    phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                      rate_continuity=True, rate_continuity_scaler=100.0,
                      rate2_continuity=False, targets=['alpha'])

    phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
    phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
    phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

    phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
    phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
    phase.add_boundary_constraint('gam', loc='final', equals=0.0)

    if path_constraints:
        phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
        phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

    # Unnecessary but included to test capability
    if path_constraints:
        phase.add_path_constraint(name='alpha', lower=-8, upper=8)
        phase.add_path_constraint(name=f'{time_name}', lower=0, upper=400)
        phase.add_path_constraint(name=f'{time_name}_phase', lower=0, upper=400)

    # Minimize time at the end of the phase
    phase.add_objective(time_name, loc='final', ref=1.0)

    # test mixing wildcard ODE variable expansion and unit overrides
    phase.add_timeseries_output(['aero.*', 'prop.thrust', 'prop.m_dot'],
                                units={'aero.f_lift': 'lbf', 'prop.thrust': 'lbf'})

    # test adding rate as timeseries output
    if add_rate:
        phase.add_timeseries_rate_output('aero.mach')

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=force_alloc_complex)

    phase.set_time_val(initial=0.0, duration=350.0)
    phase.set_state_val('r', [0.0, 111319.54])
    phase.set_state_val('h', [100.0, 20000.0])
    phase.set_state_val('v', [135.964, 283.159])
    phase.set_state_val('gam', [0.0, 0.0])
    phase.set_state_val('m', [19030.468, 16841.431])
    phase.set_control_val('alpha', [0.0, 0.0])

    if transcription == 'birkhoff':
        phase.set_simulate_options(times_per_seg=200, atol=1.0E-6, rtol=1.0E-6)

    if transcription == 'birkhoff':
        phase.set_simulate_options(times_per_seg=200, atol=1.0E-6, rtol=1.0E-6)

    p.run_model()
    with open('check_partials.txt', 'w') as f:
        with np.printoptions(linewidth=1024, edgeitems=1024):
            p.check_partials(compact_print=False, method='cs', show_only_incorrect=True, out_stream=f)

    dm.run_problem(p, simulate=simulate, make_plots=make_plots, plot_kwargs={'x_name': time_name})

    return p


@use_tempdirs
class TestMinTimeClimb(unittest.TestCase):

    def _test_results(self, p, time_name='time'):
        """ Verify the results of the optimization. """
        # Verify that ODE output mach is added to the timeseries
        assert_near_equal(p.get_val('traj.phase0.timeseries.mach')[-1], 1.0, tolerance=1.0E-2)

        # Check that time matches to within 1% of an externally verified solution.
        assert_near_equal(p.get_val(f'traj.phase0.timeseries.{time_name}')[-1], 321.0, tolerance=0.02)

    def _test_wilcard_outputs(self, p):
        """ Test that all wilcard outputs are provided. """
        output_dict = get_promoted_vars(p.model, iotypes=('output',))
        ts = {k: v for k, v in output_dict.items() if 'timeseries.' in k}
        for c in ['mach', 'CD0', 'kappa', 'CLa', 'CL', 'CD', 'q', 'f_lift', 'f_drag', 'thrust', 'm_dot']:
            assert any([True for t in ts if 'timeseries.' + c in t])

    def _test_timeseries_units(self, p):
        """ Test that the units from the timeseries are correct. """
        output_dict = get_promoted_vars(p.model, iotypes=('output',))
        assert output_dict['traj.phase0.timeseries.thrust']['units'] == 'lbf'  # no wildcard, from units dict
        assert output_dict['traj.phase0.timeseries.m_dot']['units'] == 'kg/s'  # no wildcard, from ODE
        assert output_dict['traj.phase0.timeseries.f_drag']['units'] == 'N'    # wildcard, from ODE
        assert output_dict['traj.phase0.timeseries.f_lift']['units'] == 'lbf'  # wildcard, from units dict

    def _test_mach_rate(self, p, plot=False, time_name='time'):
        """ Test that the mach rate is provided by the timeseries and is accurate. """
        # Verify correct timeseries output of mach_rate
        output_dict = get_promoted_vars(p.model, iotypes=('output',))
        ts = {k: v for k, v in output_dict.items() if 'timeseries.' in k}
        self.assertTrue('traj.phase0.timeseries.mach_rate' in ts)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        case = om.CaseReader(sol_db).get_case('final')
        sim_case = om.CaseReader(sol_db).get_case('final')

        time = case[f'traj.phase0.timeseries.{time_name}'][:, 0]
        mach = case['traj.phase0.timeseries.mach'][:, 0]
        mach_rate = case['traj.phase0.timeseries.mach_rate'][:, 0]

        sim_time = sim_case[f'traj.phase0.timeseries.{time_name}'][:, 0]
        sim_mach = sim_case['traj.phase0.timeseries.mach'][:, 0]
        sim_mach_rate = sim_case['traj.phase0.timeseries.mach_rate'][:, 0]

        # Fit a numpy polynomial segment by segment to mach vs time, and compare the derivatives to mach_rate
        gd = p.model.traj.phases.phase0.options['transcription'].grid_data
        seg_idxs = gd.subset_segment_indices['all']
        num_seg = seg_idxs.shape[0]

        sim_seg_idxs = np.zeros_like(seg_idxs)
        sim_seg_idxs[:, 0] = np.arange(0, 10*num_seg, 10, dtype=int)
        sim_seg_idxs[:, 1] = sim_seg_idxs[:, 0] + 10

        if plot:
            fig, axes = plt.subplots(2, 1, sharex=True)
            color = iter(cm.viridis(np.linspace(0, 1, num_seg)))
            axes[0].set_ylabel('Mach')
            axes[1].set_ylabel('Mach rate (1/s)')
            axes[1].set_xlabel('time (s)')

        for i in range(num_seg):
            time_seg = time[seg_idxs[i, 0]: seg_idxs[i, 1]]
            mach_seg = mach[seg_idxs[i, 0]: seg_idxs[i, 1]]
            order = len(time_seg) - 1
            p = P.fit(time_seg, mach_seg, order)
            deriv = p.deriv(1)

            time_nodes = time[seg_idxs[i, 0]: seg_idxs[i, 1]]
            mach_nodes = mach[seg_idxs[i, 0]: seg_idxs[i, 1]]
            mach_rate_nodes = mach_rate[seg_idxs[i, 0]: seg_idxs[i, 1]]

            if plot:
                c = next(color)
                t_plot = np.linspace(time_seg[0], time_seg[-1], 20)
                m_plot = p(t_plot)
                m_rate_plot = deriv(t_plot)
                axes[0].plot(t_plot, m_plot, c=c)
                axes[1].plot(t_plot, m_rate_plot, '--', c=c)

                axes[0].plot(time[seg_idxs[i, 0]: seg_idxs[i, 1]], mach[seg_idxs[i, 0]: seg_idxs[i, 1]], 'o', c=c)
                axes[1].plot(time[seg_idxs[i, 0]: seg_idxs[i, 1]], mach_rate[seg_idxs[i, 0]: seg_idxs[i, 1]], 'o', c=c)

                axes[0].plot(sim_time[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]],
                             sim_mach[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]], ':', c=c, lw=2)
                axes[1].plot(sim_time[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]],
                             sim_mach_rate[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]], ':', c=c, lw=2)

            assert_near_equal(mach_nodes, p(time_nodes), tolerance=1.0E-9)
            assert_near_equal(mach_rate_nodes, deriv(time_nodes), tolerance=1.0E-9)

        # Comparing the mach rate over the entire trajectory since it is expected to be off at some points due to
        # the equidistant time-spacing of nodes in ExplicitShooting's timeseries outputs.
        assert_timeseries_near_equal(t_ref=time, x_ref=mach_rate, t_check=sim_time, x_check=sim_mach_rate,
                                     abs_tolerance=0.02, rel_tolerance=0.02)

        if plot:
            plt.show()

    @require_pyoptsparse(optimizer='IPOPT')
    def test_results_gauss_lobatto(self):
        NUM_SEG = 12
        ORDER = 3
        p = min_time_climb(optimizer='IPOPT', num_seg=NUM_SEG, transcription_order=ORDER,
                           transcription='gauss-lobatto', add_rate=True)

        self._test_results(p)

        self._test_wilcard_outputs(p)

        self._test_timeseries_units(p)

        self._test_mach_rate(p)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_results_radau(self):
        NUM_SEG = 15
        ORDER = 3
        p = min_time_climb(optimizer='IPOPT', num_seg=NUM_SEG, transcription_order=ORDER,
                           transcription='radau-ps', add_rate=True)

        self._test_results(p)

        self._test_wilcard_outputs(p)

        self._test_timeseries_units(p)

        self._test_mach_rate(p, plot=False)

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(om_version()[0] < (3, 32, 2), 'Test requires OpenMDAO 3.32.2 or later')
    def test_results_birkhoff(self):
        NUM_SEG = 1
        ORDER = 30
        p = min_time_climb(optimizer='IPOPT', num_seg=NUM_SEG, transcription_order=ORDER,
                           transcription='birkhoff', add_rate=False, simulate=True, path_constraints=True,
                           force_alloc_complex=True)

        self._test_results(p)

        self._test_wilcard_outputs(p)

        self._test_timeseries_units(p)


@use_tempdirs
class TestMinTimeClimbWithReports(TestMinTimeClimb):

    def setUp(self):
        self.testflo_running = os.environ.pop('TESTFLO_RUNNING', None)

    def tearDown(self):
        # restore what was there before running the test
        if self.testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = self.testflo_running

    def _test_traj_results_report(self, p):
        html_file = _get_reports_dir(p) / 'traj_results_report.html'
        self.assertTrue(html_file.exists(), msg=f'{html_file} does not exist!')

        with open(html_file) as f:
            html_data = f.read()

        expected_labels = ['"axis_label":"alpha (deg)"',
                           '"axis_label":"t (s)"',
                           '"axis_label":"CD (None)"',
                           '"axis_label":"CD0 (None)"',
                           '"axis_label":"CL (None)"',
                           '"axis_label":"CLa (None)"',
                           '"axis_label":"f_drag (N)"',
                           '"axis_label":"f_lift (lbf)"',
                           '"axis_label":"gam (rad)"',
                           '"axis_label":"h (m)"',
                           '"axis_label":"kappa (None)"',
                           '"axis_label":"m (kg)"',
                           '"axis_label":"m_dot (kg/s)"',
                           '"axis_label":"mach (None)"',
                           '"axis_label":"mach_rate (None)"',
                           '"axis_label":"q (N/m**2)"',
                           '"axis_label":"r (m)"',
                           '"axis_label":"thrust (lbf)"',
                           '"axis_label":"v (m/s)"']

        for label in expected_labels:
            self.assertIn(label, html_data)

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(not bokeh_available, 'bokeh is not available')
    def test_results_gauss_lobatto_renamed_time(self):
        with set_env_vars_context(OPENMDAO_REPORTS='1'):
            with dm.options.temporary(plots='bokeh'):
                NUM_SEG = 12
                ORDER = 3
                p = min_time_climb(optimizer='IPOPT', num_seg=NUM_SEG, transcription_order=ORDER,
                                   force_alloc_complex=True,
                                   transcription='gauss-lobatto', add_rate=True, time_name='t',
                                   make_plots=True)

                self._test_results(p, time_name='t')

                self._test_wilcard_outputs(p)

                self._test_timeseries_units(p)

                self._test_mach_rate(p, time_name='t')

                self._test_traj_results_report(p)

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(not bokeh_available, 'bokeh is not available')
    def test_results_radau_renamed_time(self):
        with set_env_vars_context(OPENMDAO_REPORTS='1'):
            with dm.options.temporary(plots='bokeh'):
                NUM_SEG = 15
                ORDER = 3
                p = min_time_climb(optimizer='IPOPT', num_seg=NUM_SEG, transcription_order=ORDER,
                                   transcription='radau-ps', add_rate=True, time_name='t',
                                   force_alloc_complex=True, make_plots=True)

                self._test_results(p, time_name='t')

                self._test_wilcard_outputs(p)

                self._test_timeseries_units(p)

                self._test_mach_rate(p, plot=False, time_name='t')

                self._test_traj_results_report(p)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm

from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


class ParamComp(om.ExplicitComponent):

    def setup(self):

        self.add_input('falling_object:mass', np.array([0.3, 0.5, 0.9]), units='kg')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        pass


class ODEGroup(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem('p1', ParamComp(), promotes=['*'])
        self.add_subsystem('brach', BrachistochroneODE(num_nodes=nn), promotes=['*'])


def make_traj(transcription='gauss-lobatto', transcription_order=3, compressed=False,
              connected=False, param_mode='param_sequence'):

    t = {'gauss-lobatto': dm.GaussLobatto(num_segments=5, order=transcription_order, compressed=compressed),
         'radau': dm.Radau(num_segments=20, order=transcription_order, compressed=compressed)}

    traj = dm.Trajectory()

    if param_mode == 'param_sequence':
        traj.add_parameter('c', opt=False, val=1.5, units='DU/TU',
                           targets={'burn1': ['c'], 'coast': ['c'], 'burn2': ['c']})
    elif param_mode == 'param_sequence_missing_phase':
        traj.add_parameter('c', opt=False, val=1.5, units='DU/TU',
                           targets={'burn1': ['c'], 'burn2': ['c']})
    elif param_mode == 'param_no_targets':
        traj.add_parameter('c', val=1.5, units='DU/TU')

    # First Phase (burn)

    burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    burn1 = traj.add_phase('burn1', burn1)

    burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
    burn1.add_state('r', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='r_dot', targets=['r'], units='DU')
    burn1.add_state('theta', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='theta_dot', targets=['theta'], units='rad')
    burn1.add_state('vr', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='vr_dot', targets=['vr'], units='DU/TU')
    burn1.add_state('vt', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='vt_dot', targets=['vt'], units='DU/TU')
    burn1.add_state('accel', fix_initial=True, fix_final=False,
                    rate_source='at_dot', targets=['accel'], units='DU/TU**2')
    burn1.add_state('deltav', fix_initial=True, fix_final=False,
                    rate_source='deltav_dot', targets=None, units='DU/TU')
    burn1.add_control('u1', rate_continuity=True, rate2_continuity=True, units='deg', scaler=0.01,
                      rate_continuity_scaler=0.001, rate2_continuity_scaler=0.001,
                      lower=-30, upper=30, targets=['u1'])
    # Second Phase (Coast)
    coast = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 50), duration_ref=50, units='TU')
    coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='r_dot', targets=['r'], units='DU')
    coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='theta_dot', targets=['theta'], units='rad')
    coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='vr_dot', targets=['vr'], units='DU/TU')
    coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='vt_dot', targets=['vt'], units='DU/TU')
    coast.add_state('accel', fix_initial=True, fix_final=True,
                    rate_source='at_dot', targets=['accel'], units='DU/TU**2')
    coast.add_state('deltav', fix_initial=False, fix_final=False,
                    rate_source='deltav_dot', units='DU/TU', targets=None)

    coast.add_parameter('u1', opt=False, val=0.0, units='deg', targets=['u1'])

    # Third Phase (burn)
    burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    traj.add_phase('coast', coast)
    traj.add_phase('burn2', burn2)

    burn2.set_time_options(initial_bounds=(0.5, 50), duration_bounds=(.5, 10), initial_ref=10, units='TU')
    burn2.add_state('r', fix_initial=False, fix_final=True, defect_scaler=100.0,
                    rate_source='r_dot', targets=['r'], units='DU')
    burn2.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='theta_dot', targets=['theta'], units='rad')
    burn2.add_state('vr', fix_initial=False, fix_final=True, defect_scaler=1000.0,
                    rate_source='vr_dot', targets=['vr'], units='DU/TU')
    burn2.add_state('vt', fix_initial=False, fix_final=True, defect_scaler=1000.0,
                    rate_source='vt_dot', targets=['vt'], units='DU/TU')
    burn2.add_state('accel', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='at_dot', targets=['accel'], units='DU/TU**2')
    burn2.add_state('deltav', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='deltav_dot', units='DU/TU', targets=None)

    burn2.add_objective('deltav', loc='final', scaler=100.0)

    burn2.add_control('u1', rate_continuity=True, rate2_continuity=True, units='deg',
                      scaler=0.01, lower=-180, upper=180, targets=['u1'])

    if 'sequence_missing_phase' in param_mode:
        coast.add_parameter('c', val=0.0, units='DU/TU', targets=['c'])
    elif 'no_targets' in param_mode:
        burn1.add_parameter('c', val=0.0, units='DU/TU', targets=['c'])
        coast.add_parameter('c', val=0.0, units='DU/TU', targets=['c'])
        burn2.add_parameter('c', val=0.0, units='DU/TU', targets=['c'])

    burn1.add_timeseries_output('pos_x')
    coast.add_timeseries_output('pos_x')
    burn2.add_timeseries_output('pos_x')

    burn1.add_timeseries_output('pos_y')
    coast.add_timeseries_output('pos_y')
    burn2.add_timeseries_output('pos_y')

    # Link Phases
    if connected:
        traj.link_phases(phases=['burn1', 'coast'],
                         vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'],
                         connected=True)

        # No direct connections to the end of a phase.
        traj.link_phases(phases=['burn2', 'coast'],
                         vars=['r', 'theta', 'vr', 'vt', 'deltav'],
                         locs=('final', 'final'))
        traj.link_phases(phases=['burn2', 'coast'],
                         vars=['time'], locs=('final', 'final'))

        traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'],
                         locs=('final', 'final'))

    else:
        traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                         vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])

        traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

    return traj


def two_burn_orbit_raise_problem(transcription='gauss-lobatto', optimizer='SLSQP', r_target=3.0,
                                 transcription_order=3, compressed=False, run_driver=True, simulate=True,
                                 show_output=True, connected=False, param_mode='param_sequence'):

    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 100
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        if show_output:
            p.driver.opt_settings['iSumm'] = 6
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['linear_solver'] = 'mumps'
        p.driver.opt_settings['mu_strategy'] = 'adaptive'
        # p.driver.opt_settings['derivative_test'] = 'first-order'

    traj = make_traj(transcription=transcription, transcription_order=transcription_order,
                     compressed=compressed, connected=connected, param_mode=param_mode)
    p.model.add_subsystem('orbit_transfer', subsys=traj)

    # Finish Problem Setup

    # Needed to move the direct solver down into the phases for use with MPI.
    #  - After moving down, used fewer iterations (about 30 less)

    p.driver.add_recorder(om.SqliteRecorder('two_burn_orbit_raise_example.db'))

    p.setup(check=True)

    # Set Initial Guesses
    p.set_val('orbit_transfer.parameters:c', val=1.5, units='DU/TU')

    burn1 = p.model.orbit_transfer.phases.burn1
    burn2 = p.model.orbit_transfer.phases.burn2
    coast = p.model.orbit_transfer.phases.coast

    if burn1 in p.model.orbit_transfer.phases._subsystems_myproc:
        p.set_val('orbit_transfer.burn1.t_initial', val=0.0)
        p.set_val('orbit_transfer.burn1.t_duration', val=2.25)
        p.set_val('orbit_transfer.burn1.states:r', val=burn1.interp('r', [1, 1.5]))
        p.set_val('orbit_transfer.burn1.states:theta', val=burn1.interp('theta', [0, 1.7]))
        p.set_val('orbit_transfer.burn1.states:vr', val=burn1.interp('vr', [0, 0]))
        p.set_val('orbit_transfer.burn1.states:vt', val=burn1.interp('vt', [1, 1]))
        p.set_val('orbit_transfer.burn1.states:accel', val=burn1.interp('accel', [0.1, 0]))
        p.set_val('orbit_transfer.burn1.states:deltav', val=burn1.interp('deltav', [0, 0.1]))
        p.set_val('orbit_transfer.burn1.controls:u1', val=burn1.interp('u1', [-3.5, 13.0]))

    if coast in p.model.orbit_transfer.phases._subsystems_myproc:
        p.set_val('orbit_transfer.coast.t_initial', val=2.25)
        p.set_val('orbit_transfer.coast.t_duration', val=3.0)

        p.set_val('orbit_transfer.coast.states:r', val=coast.interp('r', [1.3, 1.5]))
        p.set_val('orbit_transfer.coast.states:theta', val=coast.interp('theta', [2.1767, 1.7]))
        p.set_val('orbit_transfer.coast.states:vr', val=coast.interp('vr', [0.3285, 0]))
        p.set_val('orbit_transfer.coast.states:vt', val=coast.interp('vt', [0.97, 1]))
        p.set_val('orbit_transfer.coast.states:accel', val=coast.interp('accel', [0, 0]))

    if burn2 in p.model.orbit_transfer.phases._subsystems_myproc:
        p.set_val('orbit_transfer.burn2.t_initial', val=5.25)
        p.set_val('orbit_transfer.burn2.t_duration', val=1.75)

        p.set_val('orbit_transfer.burn2.states:r', val=burn2.interp('r', [1, r_target]))
        p.set_val('orbit_transfer.burn2.states:theta', val=burn2.interp('theta', [0, 4.0]))
        p.set_val('orbit_transfer.burn2.states:vr', val=burn2.interp('vr', [0, 0]))
        p.set_val('orbit_transfer.burn2.states:vt', val=burn2.interp('vt', [1, np.sqrt(1 / r_target)]))
        p.set_val('orbit_transfer.burn2.states:deltav', val=burn2.interp('deltav', [0.1, 0.2]))
        p.set_val('orbit_transfer.burn2.states:accel', val=burn2.interp('accel', [0.1, 0]))

        p.set_val('orbit_transfer.burn2.controls:u1', val=burn2.interp('u1', [0, 0]))

    dm.run_problem(p, run_driver=run_driver, simulate=simulate)

    return p


@use_tempdirs
class TestTrajectoryParameters(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_param_explicit_connections_to_sequence(self):
        """
        Test that, when setting up a trajectory parameter, we can explicitly provide a sequence
        in each phase as targets and a corresponding parameter for the phase will
        automatically be added.
        """
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer='IPOPT', run_driver=False,
                                         simulate=False, show_output=False, param_mode='param_sequence')

        traj_c = p.get_val('orbit_transfer.parameter_vals:c')
        burn1_c = p.get_val('orbit_transfer.burn1.parameter_vals:c')
        coast_c = p.get_val('orbit_transfer.coast.parameter_vals:c')
        burn2_c = p.get_val('orbit_transfer.burn2.parameter_vals:c')

        assert_near_equal(burn1_c, traj_c)
        assert_near_equal(coast_c, traj_c)
        assert_near_equal(burn2_c, traj_c)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_param_explicit_connections_to_sequence_missing_phase(self):
        """
        Test that, when setting up a trajectory parameter with a phase omitted from input,
        that we attempt to connect to an existing input variable in that phase of the same name.
        """
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer='IPOPT', run_driver=False, simulate=False,
                                         show_output=False,
                                         param_mode='param_sequence_missing_phase')

        traj_c = p.get_val('orbit_transfer.parameter_vals:c')
        burn1_c = p.get_val('orbit_transfer.burn1.parameter_vals:c')
        coast_c = p.get_val('orbit_transfer.coast.parameter_vals:c')
        burn2_c = p.get_val('orbit_transfer.burn2.parameter_vals:c')

        assert_near_equal(burn1_c, traj_c)
        assert_near_equal(coast_c, traj_c)
        assert_near_equal(burn2_c, traj_c)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_param_no_targets(self):
        """
        Test that, when setting up a trajectory parameter with a phase omitted from input,
        that we attempt to connect to an existing input variable in that phase of the same name.
        """
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer='IPOPT', run_driver=False,
                                         show_output=False, simulate=False,
                                         param_mode='param_no_targets')

        traj_c = p.get_val('orbit_transfer.parameter_vals:c')
        burn1_c = p.get_val('orbit_transfer.burn1.parameter_vals:c')
        coast_c = p.get_val('orbit_transfer.coast.parameter_vals:c')
        burn2_c = p.get_val('orbit_transfer.burn2.parameter_vals:c')

        assert_near_equal(burn1_c, traj_c)
        assert_near_equal(coast_c, traj_c)
        assert_near_equal(burn2_c, traj_c)


@use_tempdirs
class TestPhaseParameterPromotion(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_promotes_parameter(self):

        for transcription in ['radau-ps', 'gauss-lobatto']:
            with self.subTest(msg=transcription):
                optimizer = 'SLSQP'
                num_segments = 10
                transcription_order = 3
                compressed = False

                p = om.Problem(model=om.Group())

                p.driver = om.pyOptSparseDriver()
                OPT, OPTIMIZER = set_pyoptsparse_opt(optimizer, fallback=True)
                p.driver.options['optimizer'] = OPTIMIZER
                p.driver.declare_coloring()

                if transcription == 'gauss-lobatto':
                    t = dm.GaussLobatto(num_segments=num_segments,
                                        order=transcription_order,
                                        compressed=compressed)
                elif transcription == 'radau-ps':
                    t = dm.Radau(num_segments=num_segments,
                                 order=transcription_order,
                                 compressed=compressed)

                traj = dm.Trajectory()
                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

                traj.add_phase('phase0', phase, promotes_inputs=['t_initial', 't_duration', 'parameters:g'])

                p.model.add_subsystem('traj', traj)

                phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

                phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=False,
                                units='m', rate_source='xdot')
                phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=False,
                                units='m', rate_source='ydot')
                phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=False,
                                units='m/s', rate_source='vdot', targets=['v'])

                phase.add_control('theta', continuity=True, rate_continuity=True,
                                  units='deg', lower=0.01, upper=179.9, targets=['theta'])

                phase.add_parameter('g', units='m/s**2', val=9.80665, targets=['g'])

                phase.add_boundary_constraint('x', loc='final', equals=10)
                phase.add_boundary_constraint('y', loc='final', equals=5)
                # Minimize time at the end of the phase
                phase.add_objective('time_phase', loc='final', scaler=10)

                p.model.linear_solver = om.DirectSolver()
                p.setup(check=True)

                p.set_val('traj.t_initial', 0.0)
                p.set_val('traj.t_duration', 2.0)

                p.set_val('traj.phase0.states:x', phase.interp('x', ys=[0, 10]))
                p.set_val('traj.phase0.states:y', phase.interp('y', ys=[10, 5]))
                p.set_val('traj.phase0.states:v', phase.interp('v', ys=[0, 9.9]))
                p.set_val('traj.phase0.controls:theta', phase.interp('theta', ys=[5, 100]))
                p.set_val('traj.parameters:g', 9.80665)

                p.run_driver()

                assert_near_equal(p['traj.t_duration'], 1.8016, tolerance=1.0E-4)


@use_tempdirs
class TestTrajectoryParametersDefinition(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_parameter_shape_definition(self):

        for transcription in ['radau-ps', 'gauss-lobatto']:
            with self.subTest(msg=transcription):
                optimizer = 'SLSQP'
                num_segments = 10
                transcription_order = 3
                compressed = False

                p = om.Problem(model=om.Group())

                p.driver = om.pyOptSparseDriver(print_results=False)
                OPT, OPTIMIZER = set_pyoptsparse_opt(optimizer, fallback=True)
                p.driver.options['optimizer'] = OPTIMIZER
                p.driver.declare_coloring()

                if transcription == 'gauss-lobatto':
                    t = dm.GaussLobatto(num_segments=num_segments,
                                        order=transcription_order,
                                        compressed=compressed)
                elif transcription == 'radau-ps':
                    t = dm.Radau(num_segments=num_segments,
                                 order=transcription_order,
                                 compressed=compressed)

                traj = p.model.add_subsystem('traj', dm.Trajectory())
                phase = traj.add_phase('phase0',
                                       dm.Phase(ode_class=ODEGroup,
                                                transcription=t),
                                       promotes_inputs=['t_initial', 't_duration', 'parameters:g'])

                # p.model.add_subsystem('p1', ParamComp(), promotes=['*'])

                # Add a vector parameter.
                traj.add_parameter('falling_object:mass',
                                   opt=False,
                                   units='kg',
                                   shape=(3,),
                                   static_target=True,
                                   targets={'phase0': ['falling_object:mass']})

                phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

                phase.add_state('x', fix_initial=True, fix_final=True)
                phase.add_state('y', fix_initial=True, fix_final=True)
                phase.add_state('v', fix_initial=True)

                phase.add_control('theta', units='deg',
                                  rate_continuity=False, lower=0.01, upper=179.9)

                phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

                # Minimize time at the end of the phase
                phase.add_objective('time', loc='final', scaler=10)

                phase.add_path_constraint('theta_rate', lower=0, upper=100)

                p.model.linear_solver = om.DirectSolver()

                p.setup()

                phase.set_time_val(initial=0.0, duration=2.0)
                phase.set_state_val('x', [0, 10])
                phase.set_state_val('y', [10, 5])
                phase.set_state_val('v', [0, 9.9])
                phase.set_control_val('theta', [5, 100])
                phase.set_parameter_val('g', 9.80665)

                # Solve for the optimal trajectory
                p.run_driver()

                # Test the results
                assert_near_equal(p['traj.t_duration'], 1.8016, tolerance=1.0E-4)


@use_tempdirs
class TestParameterIntrospection(unittest.TestCase):

    def test_parameter_introspection_targets_none_no_valid_parameter_targets(self):
        import openmdao.api as om

        import dymos as dm

        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        p = om.Problem()

        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=5, compressed=False))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6, units='m',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0, units='m',
                        ref=100.0, defect_ref=1.0E2,
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0, units='m/s',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5, units='rad',
                        ref=1.0, defect_ref=1.0,
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5, units='kg',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False)

        # Unit introspection for phase params. This works.
        phase.add_parameter('S', val=49.2386)

        # Unit introspection for traj params. This doesn't work.
        traj.add_parameter('Isp', val=1600.0, targets=None)

        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        with self.assertRaises(ValueError) as e:
            p.setup(check=True)

        expected = 'No target was found for trajectory parameter `Isp` in any phase.\n' \
                   'Option `targets=None` but no phase in the trajectory has a parameter named `Isp`.'

        self.assertEqual(str(e.exception), expected)

    def test_parameter_introspection_targets_dict_no_valid_parameter_targets(self):
        import openmdao.api as om

        import dymos as dm

        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        p = om.Problem()

        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=5, compressed=False))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6, units='m',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0, units='m',
                        ref=100.0, defect_ref=1.0E2,
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0, units='m/s',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5, units='rad',
                        ref=1.0, defect_ref=1.0,
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5, units='kg',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False)

        # Unit introspection for phase params. This works.
        phase.add_parameter('S', val=49.2386)

        # Unit introspection for traj params. This doesn't work.
        traj.add_parameter('Isp', val=1600.0, targets={'phase0': 'Isp'})

        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        with self.assertRaises(ValueError) as e:
            p.setup(check=True)

        expected = 'Invalid target for trajectory `traj` parameter `Isp` in phase `phase0`.\n' \
                   "Target for phase `phase0` is 'Isp' but the phase has no such parameter."

        self.assertEqual(str(e.exception), expected)

    def test_parameter_introspection_targets_dict_no_valid_ode_targets(self):
        import openmdao.api as om

        import dymos as dm

        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        p = om.Problem()

        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=5, compressed=False))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6, units='m',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0, units='m',
                        ref=100.0, defect_ref=1.0E2,
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0, units='m/s',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5, units='rad',
                        ref=1.0, defect_ref=1.0,
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5, units='kg',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False)

        # Unit introspection for phase params. This works.
        phase.add_parameter('S', val=49.2386)

        # Error, no such ODE target in the phase.
        traj.add_parameter('Isp', val=1600.0, targets={'phase0': ['foo']})

        with self.assertRaises(ValueError) as e:
            p.setup(check=True)

        expected = "Invalid parameter in phase `traj.phases.phase0`.\n" \
                   "Parameter `Isp` has invalid target(s).\n" \
                   "No such ODE input: 'foo'."

        self.assertEqual(str(e.exception), expected)


if __name__ == '__main__':
    unittest.main()



import unittest
from openmdao.utils.testing_utils import use_tempdirs

try:
    import matplotlib
except ImportError:
    matplotlib = None


@use_tempdirs
class TestDocOscillator(unittest.TestCase):

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_ivp(self):
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')  # disable plotting to the screen

        from dymos.examples.oscillator.oscillator_ode import OscillatorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()

        # Instantiate a Dymos Trajectory and add it to the Problem model.
        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)

        # Instantiate a Phase and add it to the Trajectory.
        # Here the transcription is necessary but not particularly relevant.
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=4))
        traj.add_phase('phase0', phase)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('v', rate_source='v_dot', targets=['v'], units='m/s')
        phase.add_state('x', rate_source='v', targets=['x'], units='m')

        # The spring constant, damping coefficient, and mass are inputs to the system
        # that are constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

        # Setup the OpenMDAO problem
        prob.setup()

        # Assign values to the times and states
        phase.set_time_val(initial=0.0, duration=15.0)
        phase.set_state_val('x', 10.0)
        phase.set_state_val('v', 0.0)
        phase.set_parameter_val('k', 1.0)
        phase.set_parameter_val('c', 0.5)
        phase.set_parameter_val('m', 1.0)

        # Perform a single execution of the model (executing the model is required before simulation).
        prob.run_model()

        # Perform an explicit simulation of our ODE from the initial conditions.
        sim_out = traj.simulate(times_per_seg=50)

        # Plot the state values obtained from the phase timeseries objects in the simulation output.
        t_sol = prob.get_val('traj.phase0.timeseries.time')
        t_sim = sim_out.get_val('traj.phase0.timeseries.time')

        states = ['x', 'v']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            sol = axes[i].plot(t_sol, prob.get_val(f'traj.phase0.timeseries.{state}'), 'o')
            sim = axes[i].plot(t_sim, sim_out.get_val(f'traj.phase0.timeseries.{state}'), '-')
            axes[i].set_ylabel(state)
        axes[-1].set_xlabel('time (s)')
        fig.legend((sol[0], sim[0]), ('solution', 'simulation'), loc='lower right', ncol=2)
        plt.tight_layout()
        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_ivp_solver(self):
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')  # disable plotting to the screen

        from dymos.examples.oscillator.oscillator_ode import OscillatorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()

        # Instantiate a Dymos Trajectory and add it to the Problem model.
        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)

        # Instantiate a Phase and add it to the Trajectory.
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=4, solve_segments='forward'))
        traj.add_phase('phase0', phase)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('x', fix_initial=True, rate_source='v', targets=['x'], units='m')
        phase.add_state('v', fix_initial=True, rate_source='v_dot', targets=['v'], units='m/s')

        # The spring constant, damping coefficient, and mass are inputs to the system that are
        # constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

        # Setup the OpenMDAO problem
        prob.setup()

        # Assign values to the times and states
        phase.set_time_val(initial=0.0, duration=15.0)
        phase.set_state_val('x', 10.0)
        phase.set_state_val('v', 0.0)
        phase.set_parameter_val('k', 1.0)
        phase.set_parameter_val('c', 0.5)
        phase.set_parameter_val('m', 1.0)

        # Now converge the dynamics using a nonlinear solver to find the trajectory without optimization.
        prob.run_model()

        # Perform an explicit simulation of our ODE from the initial conditions.
        sim_out = traj.simulate(times_per_seg=50)

        # Plot the state values obtained from the phase timeseries objects in the simulation output.
        t_sol = prob.get_val('traj.phase0.timeseries.time')
        t_sim = sim_out.get_val('traj.phase0.timeseries.time')

        states = ['x', 'v']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            sol = axes[i].plot(t_sol, prob.get_val(f'traj.phase0.timeseries.{state}'), 'o')
            sim = axes[i].plot(t_sim, sim_out.get_val(f'traj.phase0.timeseries.{state}'), '-')
            axes[i].set_ylabel(state)
        axes[-1].set_xlabel('time (s)')
        fig.legend((sol[0], sim[0]), ('solution', 'simulation'), loc='lower right', ncol=2)
        plt.tight_layout()
        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_ivp_driver(self):
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')  # disable plotting to the screen

        from dymos.examples.oscillator.oscillator_ode import OscillatorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()

        # We need an optimization driver.  To solve this simple problem ScipyOptimizerDriver will work.
        prob.driver = om.ScipyOptimizeDriver()

        # Instantiate a Dymos Trajectory and add it to the Problem model.
        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)

        # Instantiate a Phase and add it to the Trajectory.
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=4))
        traj.add_phase('phase0', phase)

        # Tell Dymos that the duration of the phase is bounded.
        phase.set_time_options(fix_initial=True, fix_duration=True)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('x', fix_initial=True, rate_source='v', targets=['x'], units='m')
        phase.add_state('v', fix_initial=True, rate_source='v_dot', targets=['v'], units='m/s')

        # The spring constant, damping coefficient, and mass are inputs to the system that are
        # constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

        # Since we're using an optimization driver, an objective is required.  We'll minimize
        # the final time in this case.
        phase.add_objective('time', loc='final')

        # Setup the OpenMDAO problem
        prob.setup()

        # Assign values to the times and states
        phase.set_time_val(initial=0.0, duration=15.0)
        phase.set_state_val('x', 10.0)
        phase.set_state_val('v', 0.0)
        phase.set_parameter_val('k', 1.0)
        phase.set_parameter_val('c', 0.5)
        phase.set_parameter_val('m', 1.0)

        # Now we're using the optimization driver to iteratively run the model and vary the
        # phase duration until the final y value is 0.
        prob.run_driver()

        # Perform an explicit simulation of our ODE from the initial conditions.
        sim_out = traj.simulate(times_per_seg=50)

        # Plot the state values obtained from the phase timeseries objects in the simulation output.
        t_sol = prob.get_val('traj.phase0.timeseries.time')
        t_sim = sim_out.get_val('traj.phase0.timeseries.time')

        states = ['x', 'v']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            sol = axes[i].plot(t_sol, prob.get_val(f'traj.phase0.timeseries.{state}'), 'o')
            sim = axes[i].plot(t_sim, sim_out.get_val(f'traj.phase0.timeseries.{state}'), '-')
            axes[i].set_ylabel(state)
        axes[-1].set_xlabel('time (s)')
        fig.legend((sol[0], sim[0]), ('solution', 'simulation'), loc='lower right', ncol=2)
        plt.tight_layout()
        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_ivp_driver_10_segs(self):
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')  # disable plotting to the screen

        from dymos.examples.oscillator.oscillator_ode import OscillatorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()

        # We need an optimization driver.  To solve this simple problem ScipyOptimizerDriver will work.
        prob.driver = om.ScipyOptimizeDriver()

        # Instantiate a Dymos Trajectory and add it to the Problem model.
        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)

        # Instantiate a Phase and add it to the Trajectory.
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=10))
        traj.add_phase('phase0', phase)

        # Tell Dymos that the duration of the phase is bounded.
        phase.set_time_options(fix_initial=True, fix_duration=True)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('x', fix_initial=True, rate_source='v', targets=['x'], units='m')
        phase.add_state('v', fix_initial=True, rate_source='v_dot', targets=['v'], units='m/s')

        # The spring constant, damping coefficient, and mass are inputs to the system that
        # are constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

        # Since we're using an optimization driver, an objective is required.  We'll minimize
        # the final time in this case.
        phase.add_objective('time', loc='final')

        # Setup the OpenMDAO problem
        prob.setup()

        # Assign values to the times and states
        phase.set_time_val(initial=0.0, duration=15.0)
        phase.set_state_val('x', 10.0)
        phase.set_state_val('v', 0.0)
        phase.set_parameter_val('k', 1.0)
        phase.set_parameter_val('c', 0.5)
        phase.set_parameter_val('m', 1.0)

        # Now we're using the optimization driver to iteratively run the model and vary the
        # phase duration until the final y value is 0.
        prob.run_driver()

        # Perform an explicit simulation of our ODE from the initial conditions.
        sim_out = traj.simulate(times_per_seg=50)

        # Plot the state values obtained from the phase timeseries objects in the simulation output.
        t_sol = prob.get_val('traj.phase0.timeseries.time')
        t_sim = sim_out.get_val('traj.phase0.timeseries.time')

        states = ['x', 'v']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            sol = axes[i].plot(t_sol, prob.get_val(f'traj.phase0.timeseries.{state}'), 'o')
            sim = axes[i].plot(t_sim, sim_out.get_val(f'traj.phase0.timeseries.{state}'), '-')
            axes[i].set_ylabel(state)
        axes[-1].set_xlabel('time (s)')
        fig.legend((sol[0], sim[0]), ('solution', 'simulation'), loc='lower right', ncol=2)
        plt.tight_layout()
        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_ivp_driver_4_segs_7_order(self):
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')  # disable plotting to the screen

        from dymos.examples.oscillator.oscillator_ode import OscillatorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()

        # We need an optimization driver.  To solve this simple problem ScipyOptimizerDriver will work.
        prob.driver = om.ScipyOptimizeDriver()

        # Instantiate a Dymos Trajectory and add it to the Problem model.
        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)

        # Instantiate a Phase and add it to the Trajectory.
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=4, order=7))
        traj.add_phase('phase0', phase)

        # Tell Dymos that the duration of the phase is bounded.
        phase.set_time_options(fix_initial=True, fix_duration=True)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('x', fix_initial=True, rate_source='v', targets=['x'], units='m')
        phase.add_state('v', fix_initial=True, rate_source='v_dot', targets=['v'], units='m/s')

        # The spring constant, damping coefficient, and mass are inputs to the system that are
        # constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

        # Since we're using an optimization driver, an objective is required.  We'll minimize
        # the final time in this case.
        phase.add_objective('time', loc='final')

        # Setup the OpenMDAO problem
        prob.setup()

        # Assign values to the times and states
        phase.set_time_val(initial=0.0, duration=15.0)
        phase.set_state_val('x', 10.0)
        phase.set_state_val('v', 0.0)
        phase.set_parameter_val('k', 1.0)
        phase.set_parameter_val('c', 0.5)
        phase.set_parameter_val('m', 1.0)

        # Now we're using the optimization driver to iteratively run the model and vary the
        # phase duration until the final y value is 0.
        prob.run_driver()

        # Perform an explicit simulation of our ODE from the initial conditions.
        sim_out = traj.simulate(times_per_seg=50)

        # Plot the state values obtained from the phase timeseries objects in the simulation output.
        t_sol = prob.get_val('traj.phase0.timeseries.time')
        t_sim = sim_out.get_val('traj.phase0.timeseries.time')

        states = ['x', 'v']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            sol = axes[i].plot(t_sol, prob.get_val(f'traj.phase0.timeseries.{state}'), 'o')
            sim = axes[i].plot(t_sim, sim_out.get_val(f'traj.phase0.timeseries.{state}'), '-')
            axes[i].set_ylabel(state)
        axes[-1].set_xlabel('time (s)')
        fig.legend((sol[0], sim[0]), ('solution', 'simulation'), loc='lower right', ncol=2)
        plt.tight_layout()
        plt.show()

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_ivp_driver_run_problem(self):
        import openmdao.api as om
        import dymos as dm
        import matplotlib.pyplot as plt
        plt.switch_backend('Agg')  # disable plotting to the screen

        from dymos.examples.oscillator.oscillator_ode import OscillatorODE

        # Instantiate an OpenMDAO Problem instance.
        prob = om.Problem()

        # We need an optimization driver.  To solve this simple problem ScipyOptimizerDriver will work.
        prob.driver = om.ScipyOptimizeDriver()

        # Instantiate a Dymos Trajectory and add it to the Problem model.
        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)

        # Instantiate a Phase and add it to the Trajectory.
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=4))
        traj.add_phase('phase0', phase)

        # Tell Dymos that the duration of the phase is bounded.
        phase.set_time_options(fix_initial=True, fix_duration=True)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('x', fix_initial=True, rate_source='v', targets=['x'], units='m')
        phase.add_state('v', fix_initial=True, rate_source='v_dot', targets=['v'], units='m/s')

        # The spring constant, damping coefficient, and mass are inputs to the system that are
        # constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

        # Since we're using an optimization driver, an objective is required.  We'll minimize
        # the final time in this case.
        phase.add_objective('time', loc='final')

        # Setup the OpenMDAO problem
        prob.setup()

        # Assign values to the times and states
        phase.set_time_val(initial=0.0, duration=15.0)
        phase.set_state_val('x', 10.0)
        phase.set_state_val('v', 0.0)
        phase.set_parameter_val('k', 1.0)
        phase.set_parameter_val('c', 0.5)
        phase.set_parameter_val('m', 1.0)

        # Now we're using the optimization driver to iteratively run the model and vary the
        # phase duration until the final y value is 0.
        dm.run_problem(prob)

        # Perform an explicit simulation of our ODE from the initial conditions.
        sim_out = traj.simulate(times_per_seg=50)

        # Plot the state values obtained from the phase timeseries objects in the simulation output.
        t_sol = prob.get_val('traj.phase0.timeseries.time')
        t_sim = sim_out.get_val('traj.phase0.timeseries.time')

        states = ['x', 'v']
        fig, axes = plt.subplots(len(states), 1)
        for i, state in enumerate(states):
            sol = axes[i].plot(t_sol, prob.get_val(f'traj.phase0.timeseries.{state}'), 'o')
            sim = axes[i].plot(t_sim, sim_out.get_val(f'traj.phase0.timeseries.{state}'), '-')
            axes[i].set_ylabel(state)
        axes[-1].set_xlabel('time (s)')
        fig.legend((sol[0], sim[0]), ('solution', 'simulation'), loc='lower right', ncol=2)
        plt.tight_layout()
        plt.show()


if __name__ == '__main__':
    unittest.main()



"""
Unit tests for tagging state rate targets in a model.
"""
import unittest
import warnings

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm


class BrachistochroneODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('static_gravity', types=(bool,), default=False,
                             desc='If True, treat gravity as a static (scalar) input, rather than '
                                  'having different values at each node.')

    def setup(self):
        nn = self.options['num_nodes']
        g_default_val = 9.80665 if self.options['static_gravity'] else 9.80665 * np.ones(nn)

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

        self.add_input('g', val=g_default_val, desc='grav. acceleration', units='m/s/s')

        self.add_input('theta', val=np.ones(nn), desc='angle of wire', units='rad')

        self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s',
                        tags=['dymos.state_rate_source:x', 'dymos.state_units:m'])

        self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s',
                        tags=['dymos.state_rate_source:y', 'dymos.state_units:m'])

        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2',
                        tags=['dymos.state_rate_source:v', 'dymos.state_units:m/s'])

        self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant',
                        units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])
        self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='xdot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='xdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='ydot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='ydot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='check', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='check', wrt='theta', rows=arange, cols=arange)

        if self.options['static_gravity']:
            c = np.zeros(self.options['num_nodes'])
            self.declare_partials(of='vdot', wrt='g', rows=arange, cols=c)
        else:
            self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['xdot'] = v * sin_theta
        outputs['ydot'] = -v * cos_theta
        outputs['check'] = v / sin_theta

    def compute_partials(self, inputs, partials):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        partials['vdot', 'g'] = cos_theta
        partials['vdot', 'theta'] = -g * sin_theta

        partials['xdot', 'v'] = sin_theta
        partials['xdot', 'theta'] = v * cos_theta

        partials['ydot', 'v'] = -cos_theta
        partials['ydot', 'theta'] = v * sin_theta

        partials['check', 'v'] = 1 / sin_theta
        partials['check', 'theta'] = -v * cos_theta / sin_theta ** 2


class ODEGroup(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        num_nodes = self.options['num_nodes']
        self.add_subsystem('ode', BrachistochroneODE(num_nodes=num_nodes),
                           promotes=['*'])


@use_tempdirs
class TestStateDiscovery(unittest.TestCase):

    def test_discovery(self):

        p = om.Problem()
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=ODEGroup,
                                        transcription=dm.GaussLobatto(num_segments=10)))

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.set_state_options('x', fix_initial=True, fix_final=True)
        phase.set_state_options('y', fix_initial=True, fix_final=True)
        phase.set_state_options('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        dm.run_problem(p)

    def test_error_messages(self):

        class BadComp(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']
                self.add_input('x', np.ones((nn, )))
                self.add_output('xdot', np.ones((nn, )), tags=['dymos.state_units:x'])

        p = om.Problem()

        phase = dm.Phase(ode_class=BadComp,
                         transcription=dm.GaussLobatto(num_segments=2))
        p.model.add_subsystem('phase', phase)

        with self.assertRaises(RuntimeError) as cm:
            p.setup()

        msg = "Error during configure_states_discovery in phase phase."
        self.assertEqual(msg, str(cm.exception))

        class BadComp(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']
                self.add_input('x', np.ones((nn, )))
                self.add_output('xdot', np.ones((nn, )), tags=['dymos.state_rate_source:x'])

        p = om.Problem()

        phase = dm.Phase(ode_class=BadComp,
                         transcription=dm.GaussLobatto(num_segments=2))
        phase.set_state_options('x', rate_source='foo')

        p.model.add_subsystem('phase', phase)

        with self.assertRaises(RuntimeError) as cm:
            p.setup()

        self.assertEqual(msg, str(cm.exception))

        class BadComp(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']
                self.add_input('x', np.ones((nn, )))
                self.add_output('xdot', np.ones((nn, )))

        p = om.Problem()

        phase = dm.Phase(ode_class=BadComp,
                         transcription=dm.GaussLobatto(num_segments=2))
        phase.add_state('x')

        p.model.add_subsystem('phase', phase)

        with self.assertRaises(RuntimeError) as cm:
            p.setup()

        self.assertEqual(msg, str(cm.exception))

    def test_deprecations(self):

        class DepComp(om.ExplicitComponent):

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']
                self.add_input('x', np.ones((nn, )), units='m')
                self.add_output('xdot', np.ones((nn, )), units='m/s',
                                tags=['state_rate_source:x', 'state_units:m'])

        p = om.Problem()

        phase = dm.Phase(ode_class=DepComp,
                         transcription=dm.GaussLobatto(num_segments=2))
        p.model.add_subsystem('phase', phase)

        expected_msg0 = "The tag 'state_rate_source:x' has a deprecated format and will no longer " \
                        "work in dymos version 2.0.0. Use 'dymos.state_rate_source:x' instead."

        expected_msg1 = "The tag 'state_units:m' has a deprecated format and will no longer work " \
                        "in dymos version 2.0.0. Use 'dymos.state_units:m' instead."

        with warnings.catch_warnings(record=True) as ctx:
            warnings.simplefilter('always')
            p.setup(check=True)

        self.assertIn(expected_msg0, [str(w.message) for w in ctx])
        self.assertIn(expected_msg1, [str(w.message) for w in ctx])


if __name__ == "__main__":
    unittest.main()
